"""
Tests image URL display field.

(Unlike test_render_ref_image, this tests rendering from saved data)

(Unlike test_upload_file, this tests rendering of a field with a direct URI value, 
not a reference to an upload/import field in some other entity)
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2015, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                    import settings
from django.contrib.auth.models     import User
from django.test                    import TestCase
from django.core.urlresolvers       import resolve, reverse
from django.test.client             import Client

from annalist                       import util
from annalist.identifiers           import ANNAL
from annalist.models.entity         import EntityRoot, Entity
from annalist.models.site           import Site
from annalist.models.collection     import Collection
from annalist.models.recordtype     import RecordType
from annalist.models.recordlist     import RecordList
from annalist.models.recordview     import RecordView
from annalist.models.recordfield    import RecordField
from annalist.models.entitytypeinfo import EntityTypeInfo

from tests                          import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )
from tests                          import init_annalist_test_site, resetSitedata
from AnnalistTestCase               import AnnalistTestCase
from entity_testutils               import (
    create_test_user,
    create_user_permissions,
    context_list_entities,
    context_list_head_fields,
    context_list_item_fields
    )
from entity_testentitydata          import (
    entity_url, entitydata_edit_url, 
    entitydata_default_view_context_data, entitydata_default_view_form_data,
    )

#   -----------------------------------------------------------------------------
#
#   Test data
#
#   -----------------------------------------------------------------------------

test_image_ref_type_create_values = (
    { 'annal:type':                 "annal:Type"
    , 'rdfs:label':                 "test_reference_type label"
    , 'rdfs:comment':               "test_reference_type comment"
    , 'annal:uri':                  "test:type/test_reference_type"
    , 'annal:type_view':            "test_reference_view"
    , 'annal:type_list':            "test_reference_list"
    })

test_image_ref_view_create_values = (
    { 'annal:type':                 "annal:View"
    , 'rdfs:label':                 "test_image_view label"
    , 'rdfs:comment':               "test_image_view comment"
    , 'annal:record_type':          ""
    , 'annal:add_field':            "yes"
    , 'annal:view_fields':
      [ { 'annal:field_id':             "Entity_id"
        , 'annal:field_placement':      "small:0,12;medium:0,6"
        }
      , { 'annal:field_id':             "Entity_label"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Entity_comment"
        , 'annal:field_placement':      "small:0,12"
        }
      , { 'annal:field_id':             "Test_image_ref"
        , 'annal:field_placement':      "small:0,12"
        }
      ]
    })

test_image_ref_field_create_values = (
    { 'annal:type':                     "annal:Field"
    , 'annal:field_name':               "ref_image"
    , 'rdfs:label':                     "test_image_ref_field label"
    , 'rdfs:comment':                   "test_image_ref_field comment"
    , 'annal:property_uri':             "test:reference"
    , 'annal:field_render_type':        "RefImage"
    , 'annal:field_value_mode':         "Value_direct"
    , 'annal:field_target_type':        "annal:Identifier"
    , 'annal:placeholder':              "(Image reference)"
    , 'annal:default_value':            ""
    })

def test_ref_entity_create_values(image_uri):
    return (
        { 'rdfs:label':                 "test_ref_image label"
        , 'rdfs:comment':               "test_ref_image comment"
        , 'test:reference':             image_uri
        })

#   -----------------------------------------------------------------------------
#
#   Linked record tests
#
#   -----------------------------------------------------------------------------

class ImageReferenceTest(AnnalistTestCase):
    """
    Tests image URI reference
    """

    def setUp(self):
        self.filepath  = "%s/README.md"%TestBaseDir
        self.fileuri   = "file://"+self.filepath
        self.imagepath = "%s/test-image.jpg"%TestBaseDir
        self.imageuri  = "file://"+self.filepath
        init_annalist_test_site()
        self.testsite    = Site(TestBaseUri, TestBaseDir)
        self.testcoll    = Collection(self.testsite, "testcoll")
        # Populate collection with linked record types, views and lists
        self.test_ref_type = RecordType.create(
            self.testcoll, "testreftype", test_image_ref_type_create_values
            )
        self.test_ref_view = RecordView.create(
            self.testcoll, "testrefview", test_image_ref_view_create_values
            )
        self.test_ref_field = RecordField.create(
            self.testcoll, "Test_image_ref", test_image_ref_field_create_values
            )
        # Create data records for testing image references:
        self.test_ref_type_info = EntityTypeInfo(
            self.testsite, self.testcoll, "testreftype", create_typedata=True
            )
        self.test_ref_type_info.create_entity("test1", test_ref_entity_create_values(self.imageuri))
        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        # resetSitedata(scope="collections")
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata()
        return

    # Utility functions

    # Tests

    def test_reference_image(self):

        # Display resource with image reference
        u = entitydata_edit_url("view", "testcoll", "testreftype", entity_id="test1", view_id="testrefview")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Check display context
        self.assertEqual(len(r.context['fields']), 4)
        i = 0
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_id")
        self.assertEqual(r.context['fields'][i].field_value,  "test1")
        i = 1
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_label")
        self.assertEqual(r.context['fields'][i].field_value,  "test_ref_image label")
        i = 2
        self.assertEqual(r.context['fields'][i].field_id,     "Entity_comment")
        self.assertEqual(r.context['fields'][i].field_value,  "test_ref_image comment")
        i = 3
        basepath = TestBasePath + "/c/testcoll/d/testreftype/"
        # print "\n*****\n"+repr(r.context['fields'][i])+"\n*****\n"
        self.assertEqual(r.context['fields'][i].field_id,           "Test_image_ref")
        self.assertEqual(r.context['fields'][i].field_value,        self.imageuri)
        self.assertEqual(r.context['fields'][i].field_value_link,   None)
        self.assertEqual(r.context['fields'][i].target_value,       self.imageuri)
        self.assertEqual(r.context['fields'][i].target_value_link,  self.imageuri)
        # Check for rendered image link
        # log.info(r.content)
        field_details = (
            { "basepath":   TestBasePath
            , "coll_id":    "testcoll"
            , "type_id":    "testupltype"
            , "entity_id":  "test1"
            , "image_uri":  self.imageuri
            , "field_id":   "ref_image"
            })
        img_element = (
            """<div class="small-12 columns"> """+
              """<div class="row view-value-row"> """+
                """<div class="view-label small-12 medium-2 columns"> """+
                  """<span>test_image_ref_field label</span> """+
                """</div> """+
                """<div class="view-value small-12 medium-10 columns"> """+
                  """<a href="%(image_uri)s" target="_blank"> """+
                    """<img src="%(image_uri)s" """+
                    """     alt="Image at '%(image_uri)s'" /> """+
                  """</a> """+
                """</div> """+
              """</div> """+
            """</div> """
            )%field_details
        self.assertContains(r, img_element, html=True)
        return

# End.



"""
Tests for repeat group rendering
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import sys
import os
import unittest
import re
from collections import OrderedDict

import logging
log = logging.getLogger(__name__)

from django.test.client                 import Client

from annalist.models.site               import Site
from annalist.models.collection         import Collection
from annalist.models.recordtype         import RecordType
from annalist.models.recordtypedata     import RecordTypeData
from annalist.models.recordview         import RecordView
from annalist.models.recordfield        import RecordField
from annalist.models.recordgroup        import RecordGroup
from annalist.models.entitydata         import EntityData

from tests                              import TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
from tests                              import init_annalist_test_site, resetSitedata
from annalist.tests.AnnalistTestCase    import AnnalistTestCase
from entity_testutils               import (
    collection_create_values,
    create_test_user
    )
from entity_testtypedata            import (
    recordtype_create_values, 
    )
from entity_testentitydata          import (
    entity_url, entitydata_edit_url, 
    default_fields
    )

class RepeatGroupRenderingTest(AnnalistTestCase):

    def setUp(self):
        # Set up basic site data
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.testcoll = Collection.create(self.testsite, "testcoll", collection_create_values("testcoll"))
        self.testtype = RecordType.create(self.testcoll, "testtype", recordtype_create_values("testtype"))
        self.testdata = RecordTypeData.create(self.testcoll, "testtype", {})
        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        # resetSitedata(scope="collections")
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata()
        return

    # Support methods

    def _create_testview(self):
        testview = RecordView.create(self.testcoll, "testview",
            { 'annal:type':         "annal:View"
            , 'annal:uri':          "test:testtype"
            , 'rdfs:label':         "Test view label"
            , 'rdfs:comment':       "Test view comment"
            , 'annal:record_type':  "testtype"
            , 'annal:add_field':    True
            , 'annal:view_fields':
              [ { 'annal:field_id':         "Entity_id"
                , 'annal:field_placement':  "small:0,12;medium:0,6"
                }
              , { 'annal:field_id':         "testrepeatfield"
                , 'annal:field_placement':  "small:0,12"
                }
              ]
            })
        self.assertTrue(testview is not None)
        return testview

    def _create_testrepeatfield(self, label_add=None, label_delete=None):
        testrepeatfield = RecordField.create(self.testcoll, "testrepeatfield",
            { "annal:type":                 "annal:Field"
            , "rdfs:label":                 "Test repeat field label"
            , "rdfs:comment":               "Test repeat field comment"
            , "annal:field_render_type":    "RepeatGroupRow"
            , "annal:field_value_type":     "annal:Field_group"
            , "annal:field_entity_type":    "test:testtype"
            , "annal:placeholder":          "(test repeat field)"
            , "annal:property_uri":         "test:repeat_fields"
            , "annal:field_placement":      "small:0,12"
            , "annal:group_ref":            "testrepeatgroup"
            , "annal:repeat_label_add":     label_add
            , "annal:repeat_label_delete":  label_delete
            })
        self.assertTrue(testrepeatfield is not None)
        return testrepeatfield

    def _create_testrepeatgroup(self):
        testrepeatgroup = RecordGroup.create(self.testcoll, "testrepeatgroup",
            { "annal:type":         "annal:Field_group"
            , "annal:uri":          "test:testrecordgroup"
            , "rdfs:label":         "Test record group label"
            , "rdfs:comment":       "Test record group comment"
            , "annal:record_type":  "test:testtype"
            , "annal:group_fields": 
                [ { "annal:field_id":   "Entity_comment" }
                ]
            })
        self.assertTrue(testrepeatgroup is not None)
        return testrepeatgroup

    def _create_testentity(self):
        testentity = EntityData.create(self.testdata, "testentity",
            { "annal:type":         "test:testtype"
            , "test:repeat_fields": []
            })
        self.assertTrue(testentity is not None)
        return testentity

    # Tests

    def test_RenderRepeatGroup(self):
        # Create view with repeat-field and repeat-group
        testview        = self._create_testview()
        testrepeatfield = self._create_testrepeatfield(
            label_add="Add group",
            label_delete="Remove group")
        testrepeatgroup = self._create_testrepeatgroup()
        # Render view
        u = entitydata_edit_url("new", "testcoll", "testtype", view_id="testview")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Test rendered values
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "testtype")
        self.assertEqual(r.context['entity_id'],        "00000001")
        self.assertEqual(r.context['action'],           "new")
        # Fields
        self.assertEqual(len(r.context['fields']), 2)
        # 1st field - Id
        self.assertEqual(r.context['fields'][0]['field_id'], 'Entity_id')
        self.assertEqual(r.context['fields'][0]['field_name'], 'entity_id')
        self.assertEqual(r.context['fields'][0]['field_label'], 'Id')
        self.assertEqual(r.context['fields'][0]['field_value'], "00000001")
        # 2nd field - repeat group
        self.assertEqual(r.context['fields'][1]['field_id'],           "testrepeatfield")
        self.assertEqual(r.context['fields'][1]['field_name'],         "testrepeatfield")
        self.assertEqual(r.context['fields'][1]['field_label'],        "Test repeat field label")
        self.assertEqual(r.context['fields'][1]['field_render_type'],  "RepeatGroupRow")
        self.assertEqual(r.context['fields'][1]['field_value_mode'],   "Value_direct")
        self.assertEqual(r.context['fields'][1]['field_target_type'],  "annal:Field_group")
        self.assertEqual(r.context['fields'][1]['field_group_ref'],    "testrepeatgroup")
        self.assertEqual(r.context['fields'][1]['group_label'],        "Test repeat field label")
        self.assertEqual(r.context['fields'][1]['group_add_label'],    "Add group")
        self.assertEqual(r.context['fields'][1]['group_delete_label'], "Remove group")
        self.assertEqual(r.context['fields'][1]['field_property_uri'], "test:repeat_fields")
        self.assertEqual(len(r.context['fields'][1]['field_value']), 0)
        self.assertEqual(r.context['fields'][1]['field_value'], "") #@@ Really?
        # Test rendered result
        field_vals = default_fields(coll_id="testcoll", type_id="testtype", entity_id="00000001")
        formrow1 = """
            <div class="small-12 medium-6 columns">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Id</span>
                </div>
                <div class="%(input_classes)s">
                    <input type="text" size="64" name="entity_id" 
                           placeholder="(entity id)" value="00000001"/>
                </div>
              </div>
            </div>
            """%field_vals(width=6)
        formrow2a = """
            <div class="%(group_label_classes)s">
              <span>Test repeat field label</span>
            </div>
            """%field_vals(width=12)
        formrow2b = """
            <div class="%(group_row_head_classes)s">
              <div class="row">
                <div class="small-1 columns">
                  &nbsp;
                </div>
                <div class="small-11 columns">
                  <div class="edit-grouprow col-head row">
                    <div class="%(col_head_classes)s">
                      <span>Comment</span>
                    </div>
                  </div>
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        formrow2c = """
            <div class="%(group_buttons_classes)s">
              <div class="row">
                <div class="small-1 columns">
                  &nbsp;
                </div>
                <div class="small-11 columns">
                  <input type="submit" name="testrepeatfield__remove" value="Remove group" />
                  <input type="submit" name="testrepeatfield__add"    value="Add group" />
                  <input type="submit" name="testrepeatfield__up"     value="Move &#x2b06;" />
                  <input type="submit" name="testrepeatfield__down"   value="Move &#x2b07;" />
                </div>
              </div>
            </div>
            """%field_vals(width=12)
        # log.info(r.content)
        self.assertContains(r, formrow1,  html=True)
        self.assertContains(r, formrow2a, html=True)
        self.assertContains(r, formrow2b, html=True)
        self.assertContains(r, formrow2c, html=True)
        return

    def test_RenderRepeatGroup_no_labels(self):
        # Create view with repeat-field and repeat-group
        testview        = self._create_testview()
        testrepeatfield = self._create_testrepeatfield(
            label_add=None,
            label_delete="")
        testrepeatgroup = self._create_testrepeatgroup()
        # Render view
        u = entitydata_edit_url("new", "testcoll", "testtype", view_id="testview")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        # Test rendered values
        self.assertEqual(r.context['coll_id'],          "testcoll")
        self.assertEqual(r.context['type_id'],          "testtype")
        self.assertEqual(r.context['entity_id'],        "00000001")
        self.assertEqual(r.context['action'],           "new")
        # Fields
        self.assertEqual(len(r.context['fields']), 2)
        # 1st field - Id
        self.assertEqual(r.context['fields'][0]['field_id'], 'Entity_id')
        self.assertEqual(r.context['fields'][0]['field_name'], 'entity_id')
        self.assertEqual(r.context['fields'][0]['field_label'], 'Id')
        self.assertEqual(r.context['fields'][0]['field_value'], "00000001")
        # 2nd field - repeat group
        self.assertEqual(r.context['fields'][1]['field_id'],           "testrepeatfield")
        self.assertEqual(r.context['fields'][1]['field_name'],         "testrepeatfield")
        self.assertEqual(r.context['fields'][1]['field_label'],        "Test repeat field label")
        self.assertEqual(r.context['fields'][1]['field_render_type'],  "RepeatGroupRow")
        self.assertEqual(r.context['fields'][1]['field_value_mode'],   "Value_direct")
        self.assertEqual(r.context['fields'][1]['field_target_type'],  "annal:Field_group")
        self.assertEqual(r.context['fields'][1]['field_group_ref'],    "testrepeatgroup")
        self.assertEqual(r.context['fields'][1]['group_label'],        "Test repeat field label")
        self.assertEqual(r.context['fields'][1]['group_add_label'],    "Add testrepeatfield")
        self.assertEqual(r.context['fields'][1]['group_delete_label'], "Remove testrepeatfield")
        self.assertEqual(r.context['fields'][1]['field_property_uri'], "test:repeat_fields")
        self.assertEqual(len(r.context['fields'][1]['field_value']), 0)
        self.assertEqual(r.context['fields'][1]['field_value'], "") #@@ Really?
        return

    def test_RenderRepeatGroup_view_no_values(self):
        # Create view with repeat-field and repeat-group
        testview        = self._create_testview()
        testrepeatfield = self._create_testrepeatfield(
            label_add="Add group",
            label_delete="Remove group")
        testrepeatgroup = self._create_testrepeatgroup()
        testentity      = self._create_testentity()
        # Render view
        u = entitydata_edit_url(
            "view", "testcoll", "testtype", entity_id="testentity", view_id="testview"
            )
        v = entity_url("testcoll", "testtype", entity_id="testentity")
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "<title>Collection testcoll</title>")
        # Test rendered result
        field_vals = default_fields(
            coll_id="testcoll", type_id="testtype", entity_id="00000001",
            view_url=v, cont_uri_param="" 
            )
        cont_uri_param = "?continuation_url="+u
        formrow1 = """
            <div class="small-12 medium-6 columns">
              <div class="row view-value-row">
                <div class="%(label_classes)s">
                  <span>Id</span>
                </div>
                <div class="%(input_classes)s">
                  <a href="%(view_url)s%(cont_uri_param)s">testentity</a>
                </div>
              </div>
            </div>
            """%field_vals(width=6)
        formrow2 = """
            <div class="row">
              <div class="%(group_label_classes)s">
                <span>Test repeat field label</span>
              </div>
              <div class="group-placeholder %(group_row_body_classes)s">
                <span>(None)</span>
              </div>
            </div>
            """%field_vals(width=12)
        # log.info(r.content)
        self.assertContains(r, formrow1,  html=True)
        self.assertContains(r, formrow2, html=True)
        return

# End.

if __name__ == "__main__":
    # import django
    # django.setup()  # Needed for template loader
    # Runtests in this module
    # runner = unittest.TextTestRunner(verbosity=2)
    # tests = unittest.TestSuite()
    # tests  = getSuite(select=sel)
    # if tests: runner.run(tests)
    unittest.main()



"""
Entity editing tests for duplicated fields
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf                    import settings
from django.db                      import models
from django.http                    import QueryDict
from django.contrib.auth.models     import User
from django.test                    import TestCase # cf. https://docs.djangoproject.com/en/dev/topics/testing/tools/#assertions
from django.test.client             import Client

from utils.SuppressLoggingContext   import SuppressLogging

from annalist.identifiers           import RDF, RDFS, ANNAL
from annalist                       import layout

from annalist.models.entitytypeinfo import EntityTypeInfo
from annalist.models.site           import Site
from annalist.models.collection     import Collection
from annalist.models.recordtype     import RecordType
from annalist.models.recordlist     import RecordList
from annalist.models.recordview     import RecordView
from annalist.models.recordfield    import RecordField
from annalist.models.recordtypedata import RecordTypeData
from annalist.models.entitydata     import EntityData

from tests                          import TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
from tests                          import init_annalist_test_site, resetSitedata
from AnnalistTestCase               import AnnalistTestCase
from entity_testutils               import (
    collection_create_values,
    continuation_url_param,
    create_test_user
    )
from entity_testtypedata            import (
    recordtype_create_values, 
    )
from entity_testviewdata            import (
    recordview_url, 
    recordview_create_values, recordview_values, recordview_values_add_field,
    recordview_entity_view_form_data
    )
from entity_testentitydata          import (
    entity_url, entitydata_edit_url, 
    entitydata_value_keys,
    entitydata_create_values, entitydata_values, entitydata_values_add_field, 
    # entitydata_default_view_form_data,
    entitydata_form_data, entitydata_form_add_field,
    default_comment
    )

#   -----------------------------------------------------------------------------
#
#   Entity edit duplicated field tests
#
#   -----------------------------------------------------------------------------

class EntityEditDupFieldTest(AnnalistTestCase):

    def setUp(self):
        init_annalist_test_site()
        self.testsite = Site(TestBaseUri, TestBaseDir)
        self.testcoll = Collection.create(self.testsite, "testcoll", collection_create_values("testcoll"))
        self.testtype = RecordType.create(self.testcoll, "testtype", recordtype_create_values("testtype"))
        self.testdata = RecordTypeData.create(self.testcoll, "testtype", {})

        # Create view with duplicate field id
        self.viewdata = recordview_create_values(view_id="DupField_view")
        recordview_values_add_field(
            self.viewdata, 
            field_id="Entity_comment", 
            field_placement="small:0,12"
            )
        recordview_values_add_field(
            self.viewdata, 
            field_id="Entity_comment",
            field_property_uri="rdfs:comment_alt",
            field_placement="small:0,12"
            )
        self.testview = RecordView.create(self.testcoll, "DupField_view", self.viewdata)

        # Login and permissions
        create_test_user(self.testcoll, "testuser", "testpassword")
        self.client = Client(HTTP_HOST=TestHost)
        loggedin = self.client.login(username="testuser", password="testpassword")
        self.assertTrue(loggedin)
        return

    def tearDown(self):
        # resetSitedata(scope="collections")
        return

    @classmethod
    def tearDownClass(cls):
        resetSitedata()
        return

    #   -----------------------------------------------------------------------------
    #   Helpers
    #   -----------------------------------------------------------------------------

    def _check_record_view_values(self, view_id, view_values):
        "Helper function checks content of record view entry with supplied view_id"
        self.assertTrue(RecordView.exists(self.testcoll, view_id))
        t = RecordView.load(self.testcoll, view_id)
        self.assertEqual(t.get_id(), view_id)
        self.assertDictionaryMatch(t.get_values(), view_values)
        return t

    def _create_entity_data(self, 
        entity_id, type_id="testtype", update="Entity", 
        comment2="Comment field 2",
        comment3="Comment field 3"
        ):
        "Helper function creates entity data with supplied entity_id"
        v = entitydata_create_values(entity_id, type_id=type_id, update=update)
        v = entitydata_values_add_field(v, "rdfs:comment", 2, comment2)
        v = entitydata_values_add_field(v, "rdfs:comment_alt", 3, comment3)
        e = EntityData.create(self.testdata, entity_id, v)
        return e    

    def _check_entity_data_values(self, 
        entity_id, type_id="testtype", update="Entity", 
        comment2="Comment field 2",
        comment3="Comment field 3"
        ):
        "Helper function checks content of form-updated record type entry with supplied entity_id"
        # log.info("_check_entity_data_values: type_id %s, entity_id %s"%(type_id, entity_id))
        typeinfo = EntityTypeInfo(self.testsite, self.testcoll, type_id)
        self.assertTrue(typeinfo.entityclass.exists(typeinfo.entityparent, entity_id))
        e = typeinfo.entityclass.load(typeinfo.entityparent, entity_id)
        self.assertEqual(e.get_id(), entity_id)
        v = entitydata_values(entity_id, type_id=type_id, update=update)
        v = entitydata_values_add_field(v, "rdfs:comment", 2, comment2)
        v = entitydata_values_add_field(v, "rdfs:comment_alt", 3, comment3)
        self.assertDictionaryMatch(e.get_values(), v)
        return e

    #   -----------------------------------------------------------------------------
    #   Form response tests
    #   -----------------------------------------------------------------------------

    def test_view_dup_field_values(self):
        self._check_record_view_values("DupField_view", self.viewdata)
        return

    def test_dup_field_display(self):
        # Create entity with duplicate fields
        self._create_entity_data("entitydupfield")
        self._check_entity_data_values(
            "entitydupfield", type_id="testtype", update="Entity", 
            comment2="Comment field 2",
            comment3="Comment field 3"
            )
        # Display entity in view with duplicate fields
        u = entitydata_edit_url(
            "edit", "testcoll", "testtype", 
            entity_id="entitydupfield",
            view_id="DupField_view"
            )
        r = self.client.get(u)
        self.assertEqual(r.status_code,   200)
        self.assertEqual(r.reason_phrase, "OK")
        self.assertContains(r, "Collection testcoll")
        # Check display context
        self.assertEqual(len(r.context['fields']), 6)
        # 4th field - 1st comment
        comment_value = "Entity coll testcoll, type testtype, entity entitydupfield"
        self.assertEqual(r.context['fields'][3]['field_id'], 'Entity_comment')
        self.assertEqual(r.context['fields'][3]['field_name'], 'Entity_comment')
        self.assertEqual(r.context['fields'][3]['field_label'], 'Comment')
        self.assertEqual(r.context['fields'][3]['field_property_uri'], "rdfs:comment")
        self.assertEqual(r.context['fields'][3]['field_placement'].field, "small-12 columns")
        self.assertEqual(r.context['fields'][3]['field_value'], comment_value)
        # 5th field - 2nd comment
        comment2_value = "Comment field 2"
        self.assertEqual(r.context['fields'][4]['field_id'], 'Entity_comment')
        self.assertEqual(r.context['fields'][4]['field_name'], 'Entity_comment__2')
        self.assertEqual(r.context['fields'][4]['field_label'], 'Comment')
        self.assertEqual(r.context['fields'][4]['field_property_uri'], "rdfs:comment__2")
        self.assertEqual(r.context['fields'][4]['field_placement'].field, "small-12 columns")
        self.assertEqual(r.context['fields'][4]['field_value'], comment2_value)
        # 6th field - 3rd comment
        comment2_value = "Comment field 3"
        self.assertEqual(r.context['fields'][5]['field_id'], 'Entity_comment')
        self.assertEqual(r.context['fields'][5]['field_name'], 'Entity_comment__3')
        self.assertEqual(r.context['fields'][5]['field_label'], 'Comment')
        self.assertEqual(r.context['fields'][5]['field_property_uri'], "rdfs:comment_alt")
        self.assertEqual(r.context['fields'][5]['field_placement'].field, "small-12 columns")
        self.assertEqual(r.context['fields'][5]['field_value'], comment2_value)
        return

    def test_dup_field_update(self):
        # Create entity with duplicate fields
        self._create_entity_data("entitydupfield")
        self._check_entity_data_values(
            "entitydupfield", type_id="testtype", update="Entity", 
            comment2="Comment field 2",
            comment3="Comment field 3"
            )
        # Post form data to update entity
        u = entitydata_edit_url(
            "edit", "testcoll", "testtype", 
            entity_id="entitydupfield",
            view_id="DupField_view"
            )
        f = entitydata_form_data(
            entity_id="entitydupfield",
            type_id="testtype",
            coll_id="testcoll", 
            action="edit", update="Updated Entity"
            )
        f = entitydata_form_add_field(f, "Entity_comment", 2, "Update comment 2")
        f = entitydata_form_add_field(f, "Entity_comment", 3, "Update comment 3")
        r = self.client.post(u, f)
        self.assertEqual(r.status_code,   302)
        self.assertEqual(r.reason_phrase, "FOUND")
        # Test resulting entity value
        self._check_entity_data_values(
            "entitydupfield", type_id="testtype", update="Updated Entity", 
            comment2="Update comment 2",
            comment3="Update comment 3"
            )
        return

# End.



"""
Annalist field group group
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import os.path
import urlparse
import shutil

import logging
log = logging.getLogger(__name__)

from django.conf import settings

from annalist                   import layout
from annalist.exceptions        import Annalist_Error
from annalist.identifiers       import ANNAL
from annalist                   import util
from annalist.models.entity     import Entity
from annalist.models.entitydata import EntityData

class RecordGroup(EntityData):

    _entitytype     = ANNAL.CURIE.Field_group
    _entitytypeid   = "_group"
    _entityview     = layout.COLL_GROUP_VIEW
    _entitypath     = layout.COLL_GROUP_PATH
    _entityaltpath  = layout.SITE_GROUP_PATH
    _entityfile     = layout.GROUP_META_FILE
    _entityref      = layout.META_GROUP_REF

    def __init__(self, parent, group_id, altparent=None):
        """
        Initialize a new RecordGroup object, without metadta (yet).

        parent      is the parent entity from which the field group is descended.
        group_id    the local identifier for the field group
        altparent   is a site object to search for this new entity,
                    allowing site-wide RecordGroup values to be found.
        """
        super(RecordGroup, self).__init__(parent, group_id, altparent)
        log.debug("RecordGroup %s: dir %s, alt %s"%(group_id, self._entitydir, self._entityaltdir))
        return

# End.



"""
Common base classes for Annalist stored entities (collections, data, metadata, etc.)

This module implements a common pattern whereby an entity is related to a parent,
with storage directories and URIs allocated by combining the parent entity and a
local identifier (slug) for the descendent.

Part of the purpose of this module is to abstract the underlying storage access
from the Annalist organization of presented entities.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import os.path
import urlparse
import shutil
import json
import errno

import logging
log = logging.getLogger(__name__)

from django.conf import settings

from annalist               import util
from annalist.exceptions    import Annalist_Error
from annalist.identifiers   import ANNAL, RDF, RDFS
from annalist.resourcetypes import file_extension, file_extension_for_content_type
from annalist.util          import make_type_entity_id

#   -------------------------------------------------------------------------------------------
#
#   EntityRoot
#
#   -------------------------------------------------------------------------------------------

class EntityRoot(object):
    """
    This is the base class for entities that do not have any parent entity
    (e.g. Annalist Site objects).

    An entity presents at least the following interface:
        cls._entitytype     type of entity (CURIE or URI)
        cls._entitytypeid   local type id (slug) used in local URI construction
        cls._entityfile     relative path to file where entity body is stored
        cls._entityref      relative reference to entity from body file
        self._entityid      ID of entity; may be None for "root" entities (e.g. site?)
        self._entityurl     URI at which entity is accessed
        # self._entityurlhost URI host at which entity is accessed (per HTTP host: header)
        # self._entityurlpath URI absolute path at which entity is accessed
        self._entitydir     directory where entity is stored
        self._values        dictionary of values in entity body
    """

    _entitytype     = ANNAL.CURIE.EntityRoot
    _entitytypeid   = None          # To be overridden
    _entityfile     = None          # To be overriden by entity subclasses..
    _entityref      = None          # Relative reference to entity from body file

    def __init__(self, entityurl, entitydir):
        """
        Initialize a new Entity object, possibly without values.  The created
        entity is not saved to disk at this stage - see .save() method.

        entityurl   is the base URI at which the entity is accessed
        entitydir   is the base directory containing the entity
        """
        self._entityid      = None
        self._entityurl     = entityurl if entityurl.endswith("/") else entityurl + "/"
        self._entitydir     = entitydir if entitydir.endswith("/") else entitydir + "/"
        self._entityalturl  = None
        self._entityaltdir  = None
        self._entityuseurl  = self._entityurl
        self._values        = None
        log.debug("EntityRoot.__init__: entity URI %s, entity dir %s"%(self._entityurl, self._entitydir))
        return

    def __repr__(self):
        return "Entity: entityid %s, values %r"%(self._entityid, self._values)

    # General entity access methods

    def set_id(self, entityid):
        self._entityid = entityid
        return

    def get_id(self):
        return self._entityid

    def get_type_id(self):
        return self._entitytypeid

    def get_type_entity_id(self):
        """
        Return type+entity Id that is unique within collection
        """
        return make_type_entity_id(self._entitytypeid, self._entityid)

    def get_label(self):
        """
        Return label string for the current entity.
        """
        return self._values.get(RDFS.CURIE.label, self._entityid)

    def get_url(self, baseurl=""):
        """
        Get fully qualified URL referred to supplied base.

        NOTE: entities are stored, as far as possible, using relative references.
        But when an absolute reference is required, the current context URL must 
        be taken into account.  If the URL returned by this function is stored, 
        subsequent references to that value will be fixed, not relative, so the value
        should only be stored where they may be used as identifiers or "permalink"
        style locators, so the data can continue to be used when moved to a new location.

        NOTE: this function always returns the primary URL associated with the entity.
        Where the entity is accessed at a secondary location, that is handled internally
        and not exposed through this function.  E.g. site-wide metadata entities are
        presented as belonging to a collection.  This allows for collection-specific
        specializations to be created without changing the URI used.
        """
        return urlparse.urljoin(baseurl, self._entityurl)

    def get_view_url(self, baseurl=""):
        """
        Return URL used to view entity data.  For metadata entities, this may be 
        different from the URL at which the resource is located, per get_url().
        The intent is to provide a URL that works regardless of whether the metadata
        is stored as site-wide or collection-specific data.

        This implementation just uses get_url(), but for entities that belong to a
        collection the URL is mapped via the web application to the underlying storage
        location.
        """
        return self.get_url(baseurl=baseurl)

    def get_view_url_path(self, baseurl=""):
        """
        Return URL path used to view entity data.  This is the URI-path of the URL
        returned by get_view_url (above)
        """
        return util.entity_url_path(self.get_view_url(), "")

    def set_values(self, values):
        """
        Set or update values for a collection
        """
        self._values = values.copy()
        self._values[ANNAL.CURIE.id]        = self._values.get(ANNAL.CURIE.id,      self._entityid)
        self._values[ANNAL.CURIE.type_id]   = self._values.get(ANNAL.CURIE.type_id, self._entitytypeid)
        self._values[ANNAL.CURIE.type]      = self._values.get(ANNAL.CURIE.type,    self._entitytype)
        if ANNAL.CURIE.url not in self._values:
            self._values[ANNAL.CURIE.url] = self.get_view_url_path()
        # log.info("set_values %r"%(self._values,))
        return self._values

    def get_values(self):
        """
        Return collection metadata values
        """
        return self._values

    def get_uri(self):
        """
        Return URI for current entity, which may be set explicitly or derived from
        its present URL.
        """
        return self._values.get(ANNAL.CURIE.uri, self._values[ANNAL.CURIE.url])

    def enum_fields(self):
        """
        Enumerate fields in entity.

        Recurses into fields that are lists or sequences of dictionaries,
        this being the structure used for repeated field groups.

        Yields `(path, value)` pairs, where `path` is a list of index values applied
        successively to access the corresponding value.
        """
        def is_dict(e):
            return isinstance(e,dict)
        def enum_f(p, d):
            for k in d:
                if isinstance(d[k], (list, tuple)) and all([is_dict(e) for e in d[k]]):
                    for i in range(len(d[k])):
                        for f in enum_f(p+[k,i], d[k][i]):
                            yield f
                else:
                    yield (p+[k], d[k])
            return
        if self._values:
            for f in enum_f([], self._values):
                yield f
        return

    def field_resource_file(self, resource_ref):
        """
        Returns a file object value for a resource associated with an entity,
        or None if the resouce is not present.
        """
        (body_dir, _) = self._dir_path()
        file_name = os.path.join(body_dir, resource_ref)
        if os.path.isfile(file_name):
            return open(file_name, "rb")
        return None

    def get_field(self, path):
        """
        Returns a field value corresponding to a path returned by enum_fields.
        """
        def get_f(p, v):
            if p == []:
                return v
            else:
                return get_f(p[1:], v[p[0]])
        return get_f(path, self._values)

    # I/O helper functions

    def _dir_path(self):
        """
        Return directory and path for current entity body file
        """
        if not self._entityfile:
            raise ValueError("Entity._dir_path without defined entity file path")
        (basedir, filepath) = util.entity_dir_path(self._entitydir, [], self._entityfile)
        return (basedir, filepath)

    def _alt_dir_path(self):
        """
        Return alternate directory and path for current entity body file
        """
        if not self._entityfile:
            raise ValueError("Entity._alt_dir_path without defined entity file path")
        if self._entityaltdir:
            (basedir, filepath) = util.entity_dir_path(self._entityaltdir, [], self._entityfile)
            return (basedir, filepath)
        return (None, None)

    def _dir_path_uri(self):
        (d, p) = self._dir_path()
        return (d, p, self._entityurl)

    def _alt_dir_path_uri(self):
        (d, p) = self._alt_dir_path()
        return (d, p, self._entityalturl)

    def _exists_path(self):
        """
        Test if the entity denoted by the current object has been created.

        If found, also sets the enity in-use URL value for .get_url()

        returns path of of object body, or None
        """
        for (d, p, u) in (self._dir_path_uri(), self._alt_dir_path_uri()):
            # log.info("_exists %s"%(p))
            if d and os.path.isdir(d):
                if p and os.path.isfile(p):
                    self._entityuseurl = u
                    return p
        return None

    def _exists(self):
        """
        Test if the entity denoted by the current object has been created.

        returns True or False.
        """
        return self._exists_path() is not None

    def _get_types(self, types):
        """
        Processes a supplied type value and returns a list of types to be stored.

        1. None is converted to an empty list
        2. A simple string is wrapped in a list
        3. A tuple is converted to a list
        4. If not already present, the current entity type is added to the list
        """
        if types is None:
            types = []
        elif isinstance(types, (tuple, list)):
            types = list(types)     # Make mutable copy
        else:
            types = [types]
        if self._entitytype not in types:
            types.append(self._entitytype)
        return types

    def _save(self):
        """
        Save current entity to Annalist storage
        """
        # @@TODO: think about capturing provenance metadata too.
        if not self._entityref:
            raise ValueError("Entity._save without defined entity reference")
        if not self._values:
            raise ValueError("Entity._save without defined entity values")
        (body_dir, body_file) = self._dir_path()
        log.debug("EntityRoot._save: dir %s, file %s"%(body_dir, body_file))
        fullpath = os.path.join(settings.BASE_DATA_DIR, "annalist_site", body_file)
        # Next is partial protection against code errors
        if not fullpath.startswith(os.path.join(settings.BASE_DATA_DIR, "annalist_site")):
            raise ValueError("Attempt to create entity file outside Annalist site tree")
        # Create directory (if needed) and save data
        util.ensure_dir(body_dir)
        values = self._values.copy()
        values['@id']   = self._entityref
        values['@type'] = self._get_types(values.get('@type', None))
        # @TODO: is this next needed?  Put logic in set_values?
        if self._entityid:
            values[ANNAL.CURIE.id] = self._entityid
        values.pop(ANNAL.CURIE.url, None)
        with open(fullpath, "wt") as entity_io:
            json.dump(values, entity_io, indent=2, separators=(',', ': '))
        self._entityuseurl  = self._entityurl
        return

    def _load_values(self):
        """
        Read current entity from Annalist storage, and return entity body.

        Adds value for 'annal:url' to the entity data returned.
        """
        body_file = self._exists_path()
        if body_file:
            try:
                with open(body_file, "r") as f:
                    entitydata = json.load(util.strip_comments(f))
                    entitydata[ANNAL.CURIE.url] = self.get_view_url_path()
                    return entitydata
            except IOError, e:
                if e.errno != errno.ENOENT:
                    raise
                log.error("EntityRoot._load_values: no file %s"%(body_file))
            except ValueError, e:
                log.error("EntityRoot._load_values: error loading %s"%(body_file))
                log.error(e)
                return (
                    { "@error": body_file 
                    , "@message": "Error loading entity values %r"%(e,)
                    })
        return None

    def _migrate_values(self, entitydata):
        """
        Default method for entity format migration hook.

        The specification for this method is that it returns an entitydata value
        which is a copy of the supplied entitydata with format migrations applied.

        This default implementation applies no migrations, and simply returns the 
        supplied value.  The method may be overridden for entity types and instances
        for which migrations are to be applied.

        NOTE:  implementations are free to apply migrations in-place.  The resulting 
        entitydata should be exctly as the supplied data *should* appear in storage
        to conform to the current format of the data.  The migration function should 
        be idempotent; i.e.
            x._migrate_values(x._migrate_values(e)) == x._migrate_values(e)
        """
        return entitydata

    def _child_dirs(self, cls, altparent):
        """
        Returns a pair of directories that may contain child entities.
        The second of the pair is descended from the "altparent", which is 
        used only for entities that are not in the main directory.

        cls         is a subclass of Entity indicating the type of children to
                    be located
        altparent   is an alternative parent entity to be checked using the specified
                    class's alternate relative path, or None if only potential children 
                    of the current entity are returned.
        """
        dir1 = os.path.dirname(os.path.join(self._entitydir, cls._entitypath or ""))
        if altparent and cls._entityaltpath:
            dir2 = os.path.dirname(os.path.join(altparent._entitydir, cls._entityaltpath))
        else:
            dir2 = None
        return (dir1, dir2)

    def _children(self, cls, altparent=None):
        """
        Iterates over candidate child identifiers that are possible instances of an 
        indicated class.  The supplied class is used to determine a subdirectory to 
        be scanned.

        cls         is a subclass of Entity indicating the type of children to
                    iterate over.
        altparent   is an alternative parent entity to be checked using the class's
                    alternate relative path, or None if only potential child IDs of the
                    current entity are returned.
        """
        coll_dir, site_dir = self._child_dirs(cls, altparent)
        assert "%" not in coll_dir, "_entitypath/_entityaltpath template variable interpolation may be in filename part only"
        site_files = []
        coll_files = []
        if site_dir and os.path.isdir(site_dir):
            site_files = os.listdir(site_dir)
        if os.path.isdir(coll_dir):
            coll_files = os.listdir(coll_dir)
        for fil in [ f for f in site_files if f not in coll_files] + coll_files:
            if util.valid_id(fil):
                yield fil
        return

    def _entity_files(self):
        """
        Iterates over files/resources (not subdirectories) that are part of the current entity.

        Returns pairs (p,f), where 'p' is a full path name, and 'f' is a filename within the 
        current entity directory. 
        """
        for f in os.listdir(self._entitydir):
            p = os.path.join(self._entitydir, f)
            if os.path.isfile(p):
                yield (p, f)
        return

    def _exists_file(self, f):
        """
        Test if a file named 'f' exists inthe current entity directory
        """
        return os.path.isfile(os.path.join(self._entitydir, f))

    def _copy_file(self, p, f):
        """
        Copy file with path 'p' to a new file 'f' in the current entity directory
        """
        new_p = os.path.join(self._entitydir, f)
        try:
            shutil.copy(p, new_p)
        except shutil.Error as e:
            log.error('shutil.copy error: %s' % e)
            return None
        except IOError as e:
            log.error('shutil.copy IOError: %s' % e.strerror)
            return None
        return new_p

    def _fileobj(self, localname, filetypeuri, mimetype, mode):
        """
        Returns a file object for accessing a blob associated with the current entity.

        localname   is the local name used to identify the file/resource among all those
                    associated with the ciurrent entity.
        filetypeuri is a URI/CURIE that identifies the type of data stored in the blob.
        mimetype    is a MIME content-type string for the resource representation used,
                    used in selecting the file extension to be used, or None in which case 
                    a default file extension for the type is used.
        mode        is a string defining how the resource is opened (using the same values
                    as the built-in `open` function).
        """
        (body_dir, body_file) = self._dir_path()  # Same as `_save`
        file_ext  = (
            file_extension_for_content_type(filetypeuri, mimetype) or 
            file_extension(filetypeuri)
            )
        file_name = os.path.join(body_dir, localname+"."+file_ext)
        return open(file_name, mode)

    # Special methods to facilitate access to entity values by dictionary operations
    # on the Entity object

    def __iter__(self):
        """
        Return entity value keys
        """
        if self._values:
            for k in self._values:
                yield k
        return

    def keys(self):
        """
        Return collection metadata value keys
        """
        return self._values.keys()

    def items(self):
        """
        Return collection metadata value fields
        """
        return self._values.items()

    def get(self, key, default):
        """
        Equivalent to dict.get() function
        """
        # log.info("entityroot.get key %r, self._values %r"%(key, self._values))
        return self[key] if self._values and key in self._values else default

    def setdefault(self, key, default):
        """
        Equivalent to dict.setdefault() function, 
        except that blank values also are overridden
        """
        if self._values and key in self._values and self._values[key]:
            result = self._values[key]
        else:
            self._values[key] = default
            result = default
        return result

    def __getitem__(self, k):
        """
        Allow direct indexing to access collection metadata value fields
        """
        return self._values[k]

    def __setitem__(self, k, v):
        """
        Allow direct indexing to update collection metadata value fields
        """
        self._values[k] = v

# End.



"""
Annalist directory/site layout
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os.path

import logging
log = logging.getLogger(__name__)

# Annalist configuration and metadata files
#
# Directory layout:
#
#   $BASE_DATA_DIR
#     annalist-site/
#       _annalist-site/
#           site_meta.json_ld
#           site_prov.json_ld
#       c/
#         <collection-id>/
#           _annalist_collection/
#               coll_meta.jsonld
#               coll_prov.jsonld
#             types/
#               <type-id>/
#                 type_meta.jsonld
#                 type_prov.jsonld
#                :
#             views/
#                 view_meta.jsonld
#                 view_prov.jsonld
#                :
#             lists/
#                 list_meta.jsonld
#                 list_prov.jsonld
#                :
#             fields/
#                 field_meta.jsonld
#                 field_prov.jsonld
#                :
#             users/  @@TBD
#               user-description
#                :
#           d/
#             <type-id>/
#               <entity-id>/
#                 entity-data.jsonld
#                 entity-prov.jsonld
#                :
#              :
#         <collection-id>/
#          :

SITE_DIR                = "annalist_site"

SITE_META_FILE          = "_annalist_site/site_meta.jsonld"
SITE_PROV_FILE          = "_annalist_site/site_prov.jsonld"
META_SITE_REF           = "../"

SITEDATA_DIR            = "_annalist_site"
SITEDATA_VIEW           = "%(id)s/"
SITEDATA_PATH           = "%(id)s"
SITEDATA_META_FILE      = "sitedata_meta.jsonld"
META_SITEDATA_REF       = "./"

SITE_COLL_VIEW          = "c/%(id)s/"
SITE_COLL_PATH          = "c/%(id)s"
COLL_META_FILE          = "_annalist_collection/coll_meta.jsonld"
COLL_PROV_FILE          = "_annalist_collection/coll_prov.jsonld"
META_COLL_REF           = "../"

COLL_TYPE_VIEW          = "d/_type/%(id)s/"
COLL_TYPE_PATH          = "_annalist_collection/types/%(id)s"
SITE_TYPE_PATH          = "_annalist_site/types/%(id)s"
TYPE_META_FILE          = "type_meta.jsonld"
TYPE_PROV_FILE          = "type_prov.jsonld"
META_TYPE_REF           = "./"

COLL_LIST_VIEW          = "d/_list/%(id)s/"
COLL_LIST_PATH          = "_annalist_collection/lists/%(id)s"
SITE_LIST_PATH          = "_annalist_site/lists/%(id)s"
LIST_META_FILE          = "list_meta.jsonld"
LIST_PROV_FILE          = "list_prov.jsonld"
META_LIST_REF           = "./"

COLL_VIEW_VIEW          = "d/_view/%(id)s/"
COLL_VIEW_PATH          = "_annalist_collection/views/%(id)s"
SITE_VIEW_PATH          = "_annalist_site/views/%(id)s"
VIEW_META_FILE          = "view_meta.jsonld"
VIEW_PROV_FILE          = "view_prov.jsonld"
META_VIEW_REF           = "./"

COLL_GROUP_VIEW         = "d/_group/%(id)s/"
COLL_GROUP_PATH         = "_annalist_collection/groups/%(id)s"
SITE_GROUP_PATH         = "_annalist_site/groups/%(id)s"
GROUP_META_FILE         = "group_meta.jsonld"
GROUP_PROV_FILE         = "group_prov.jsonld"
META_GROUP_REF          = "./"

COLL_FIELD_VIEW         = "d/_field/%(id)s/"
COLL_FIELD_PATH         = "_annalist_collection/fields/%(id)s"
SITE_FIELD_PATH         = "_annalist_site/fields/%(id)s"
FIELD_META_FILE         = "field_meta.jsonld"
FIELD_PROV_FILE         = "field_prov.jsonld"
META_FIELD_REF          = "./"

COLL_USER_VIEW         = "d/_user/%(id)s/"
COLL_USER_PATH         = "_annalist_collection/users/%(id)s"
SITE_USER_PATH         = "_annalist_site/users/%(id)s"
USER_META_FILE         = "user_meta.jsonld"
USER_PROV_FILE         = "user_prov.jsonld"
META_USER_REF          = "./"

COLL_ENUM_VIEW          = "d/%(type_id)s/%(id)s/"
COLL_ENUM_PATH          = "_annalist_collection/enums/%(type_id)s/%(id)s"
SITE_ENUM_PATH          = "_annalist_site/enums/%(type_id)s/%(id)s"
ENUM_META_FILE          = "enum_meta.jsonld"
ENUM_PROV_FILE          = "enum_prov.jsonld"
META_ENUM_REF           = "./"

COLL_TYPEDATA_VIEW      = "d/%(id)s/"
COLL_TYPEDATA_PATH      = "d/%(id)s"
TYPEDATA_META_FILE      = "type_data_meta.jsonld"
META_TYPEDATA_REF       = "./"

TYPEDATA_ENTITY_VIEW    = "%(id)s/"
TYPEDATA_ENTITY_PATH    = "%(id)s"
ENTITY_DATA_FILE        = "entity-data.jsonld"
ENTITY_PROV_FILE        = "entity-prov.jsonld"
DATA_ENTITY_REF         = "./"

# and more...

class Layout(object):
    """
    A dynamically created layout value with paths that are dynamically constructed 
    using a supplied base directory.
    """

    def __init__(self, base_data_dir):
        """
        Dynamically initialize a layout value
        """
        self.BASE_DIR       = base_data_dir
        self.SITE_DIR       = SITE_DIR
        self.SITEDATA_DIR   = SITEDATA_DIR
        self.SITE_PATH      = os.path.join(base_data_dir, SITE_DIR)
        return

# End.



"""
Annalist record type

A record type is represented by:
- an ID (slug)
- a URI
- a name/label
- a description
- ...
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import os.path
import urlparse
import shutil

import logging
log = logging.getLogger(__name__)

from django.conf import settings

from annalist                   import layout
from annalist.exceptions        import Annalist_Error
from annalist.identifiers       import ANNAL
from annalist                   import util
from annalist.models.entity     import Entity
from annalist.models.entitydata import EntityData

class RecordType(EntityData):

    _entitytype     = ANNAL.CURIE.Type
    _entitytypeid   = "_type"
    _entityview     = layout.COLL_TYPE_VIEW
    _entitypath     = layout.COLL_TYPE_PATH
    _entityaltpath  = layout.SITE_TYPE_PATH
    _entityfile     = layout.TYPE_META_FILE
    _entityref      = layout.META_TYPE_REF

    def __init__(self, parent, type_id, altparent=None):
        """
        Initialize a new RecordType object, without metadta (yet).

        parent      is the parent entity from which the type is descended.
        type_id     the local identifier for the record type
        altparent   is a site object to search for this new entity,
                    allowing site-wide RecordType values to be found.
        """
        super(RecordType, self).__init__(parent, type_id, altparent=altparent)
        log.debug("RecordType %s: dir %s, alt %s"%(type_id, self._entitydir, self._entityaltdir))
        log.debug("RecordType %s: uri %s, alt %s"%(type_id, self._entityurl, self._entityalturi))
        return

# End.



"""
Initialize Annalist server data.
"""

from __future__ import print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import sys
import logging
import importlib
import subprocess

log = logging.getLogger(__name__)

from annalist.util                  import ensure_dir
from utils.SuppressLoggingContext   import SuppressLogging

import am_errors
from am_settings                    import am_get_settings

def am_initialize(annroot, userhome, userconfig, options):
    """
    Initialize Annalist server data, database, etc.

    annroot     is the root directory for theannalist software installation.
    userhome    is the home directory for the host system user issuing the initialize command.
    userconfig  is the directory used for user-specific configuration files.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings = am_get_settings(annroot, userhome, options)
    if not settings:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) != 0:
        print("Unexpected arguments for initialize: (%s)"%(" ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    # Get config base directory from settings, and make sure it exists
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings.modulename)
    providersdir = os.path.join(sitesettings.CONFIG_BASE, "providers")
    databasedir  = os.path.dirname(sitesettings.DATABASES['default']['NAME'])
    ensure_dir(providersdir)
    ensure_dir(databasedir)
    # Initialze the database
    status = am_errors.AM_SUCCESS
    subprocess_command = "django-admin migrate --pythonpath=%s --settings=%s"%(annroot, settings.modulename)
    log.debug("am_initialize subprocess: %s"%subprocess_command)
    # OLD: status = os.system(subprocess_command)
    status = subprocess.call(subprocess_command.split())
    log.debug("am_initialize subprocess status: %s"%status)
    return status

# End.



"""
Annalist user record

A user is represented in a collectionm by:
- an ID (slug)
- a URI (currently a mailto: URI)
- a label (full name)
- a description
- a list of permissions applicable to the collection
- ...

The ID and URI must be matched by the authenticated issuer of an HTTP request for the
permissionms to be applied.  Other fields are cosmetic.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import os.path
import urlparse
import shutil

import logging
log = logging.getLogger(__name__)

from django.conf import settings

from annalist                   import layout
from annalist.exceptions        import Annalist_Error
from annalist.identifiers       import ANNAL
from annalist                   import util
from annalist.models.entity     import Entity
from annalist.models.entitydata import EntityData

class AnnalistUser(EntityData):

    _entitytype     = ANNAL.CURIE.User
    _entitytypeid   = "_user"
    _entityview     = layout.COLL_USER_VIEW
    _entitypath     = layout.COLL_USER_PATH
    _entityaltpath  = layout.SITE_USER_PATH
    _entityfile     = layout.USER_META_FILE
    _entityref      = layout.META_USER_REF

    def __init__(self, parent, type_id, altparent=None, use_altpath=False):
        """
        Initialize a new AnnalistUser object, without metadata (yet).

        parent      is the parent entity from which the type is descended.
        type_id     the local identifier for the record type
        altparent   is a site object to search for this new entity,
                    allowing site-wide AnnalistUser values to be found.
        use_altpath is set True if this entity is situated at the alternative
                    path relative to its parent.  This is used to access
                    user permissions when there is no collection context.
        """
        super(AnnalistUser, self).__init__(parent, type_id, altparent=altparent, use_altpath=use_altpath)
        log.debug("AnnalistUser %s: dir %s, alt %s"%(type_id, self._entitydir, self._entityaltdir))
        log.debug("AnnalistUser %s: uri %s, alt %s"%(type_id, self._entityurl, self._entityalturi))
        return

# End.



# HTTP session class and supporting utilites.

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2011-2013, University of Oxford"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import re   # Used for link header parsing
import httplib2
import urlparse
import logging

# Logger for this module
log = logging.getLogger(__name__)

def splitValues(txt, sep=",", lq='"<', rq='">'):
    """
    Helper function returns list of delimited values in a string,
    where delimiters in quotes are protected.

    sep is string of separator
    lq is string of opening quotes for strings within which separators are not recognized
    rq is string of corresponding closing quotes
    """
    result = []
    cursor = 0
    begseg = cursor
    while cursor < len(txt):
        if txt[cursor] in lq:
            # Skip quoted or bracketed string
            eq = rq[lq.index(txt[cursor])]  # End quote/bracket character
            cursor += 1
            while cursor < len(txt) and txt[cursor] != eq:
                if txt[cursor] == '\\': cursor += 1 # skip '\' quoted-pair
                cursor += 1
            if cursor < len(txt):
                cursor += 1 # Skip closing quote/bracket
        elif txt[cursor] in sep:
            result.append(txt[begseg:cursor])
            cursor += 1
            begseg = cursor
        else:
            cursor += 1
    # append final segment
    result.append(txt[begseg:cursor])
    return result

def testSplitValues():
    assert splitValues("a,b,c") == ['a','b','c']
    assert splitValues('a,"b,c",d') == ['a','"b,c"','d']
    assert splitValues('a, "b, c\\", c1", d') == ['a',' "b, c\\", c1"',' d']
    assert splitValues('a,"b,c",d', ";") == ['a,"b,c",d']
    assert splitValues('a;"b;c";d', ";") == ['a','"b;c"','d']
    assert splitValues('a;<b;c>;d', ";") == ['a','<b;c>','d']
    assert splitValues('"a;b";(c;d);e', ";", lq='"(', rq='")') == ['"a;b"','(c;d)','e']

def parseLinks(headerlist):
    """
    Helper function to parse 'link:' headers,
    returning a dictionary of links keyed by link relation type
    
    headerlist is a list of header (name,value) pairs
    """
    linkheaders = [ v for (h,v) in headerlist if h.lower() == "link" ]
    log.debug("parseLinks linkheaders %s"%(repr(linkheaders)))
    links = {}
    for linkheader in linkheaders:
        for linkval in splitValues(linkheader, ","):
            linkparts = splitValues(linkval, ";")
            linkmatch = re.match(r'''\s*<([^>]*)>\s*''', linkparts[0])
            if linkmatch:
                linkuri   = linkmatch.group(1)
                for linkparam in linkparts[1:]:
                    linkmatch = re.match(r'''\s*rel\s*=\s*"?(.*?)"?\s*$''', linkparam)  # .*? is non-greedy
                    if linkmatch:
                        linkrel = linkmatch.group(1)
                        log.debug("parseLinks links[%s] = %s"%(linkrel, linkuri))
                        links[linkrel] = linkuri
    return links

def testParseLinks():
    links = (
        ('Link', '<http://example.org/foo>; rel=foo'),
        ('Link', ' <http://example.org/bar> ; rel = bar '),
        ('Link', '<http://example.org/bas>; rel=bas; par = zzz , <http://example.org/bat>; rel = bat'),
        ('Link', ' <http://example.org/fie> ; par = fie '),
        ('Link', ' <http://example.org/fum> ; rel = "http://example.org/rel/fum" '),
        ('Link', ' <http://example.org/fas;far> ; rel = "http://example.org/rel/fas" '),
        )
    assert str(parseLinks(links)['foo']) == 'http://example.org/foo'
    assert str(parseLinks(links)['bar']) == 'http://example.org/bar'
    assert str(parseLinks(links)['bas']) == 'http://example.org/bas'
    assert str(parseLinks(links)['bat']) == 'http://example.org/bat'
    assert str(parseLinks(links)['http://example.org/rel/fum']) == 'http://example.org/fum'
    assert str(parseLinks(links)['http://example.org/rel/fas']) == 'http://example.org/fas;far'


# Class for exceptions raised by HTTP session

class HTTP_Error(Exception):

    def __init__(self, msg="HTTP_Error", value=None, uri=None):
        self._msg   = msg
        self._value = value
        self._uri   = uri
        return

    def __str__(self):
        txt = self._msg
        if self._uri:   txt += " for "+str(self._uri)
        if self._value: txt += ": "+repr(self._value)
        return txt

    def __repr__(self):
        return ( "HTTP_Error(%s, value=%s, uri=%s)"%
                 (repr(self._msg), repr(self._value), repr(self._uri)))


# Class for handling Access in an HTTP session

class HTTP_Session(object):
    
    """
    Client access class for HTTP session.

    Creates a session to access a single HTTP endpoint,
    and provides methods to issue requests on this session

    This class is primarily designed to access a specific endpoint, and 
    by default refuses requests for different endpoints.  But the request
    methods accept an additional "exthost" parameter that can be used to
    override this behaviour.  Specifying "exthost=True" causes the request 
    to allow URIs that use different scheme, hostname or port than the original
    request, but such requests are not issued using the access key of the HTTP
    session.
    """

    def __init__(self, baseuri, accesskey=None):
        log.debug("HTTP_Session.__init__: baseuri "+baseuri)
        self._baseuri = baseuri
        self._key     = accesskey
        parseduri     = urlparse.urlsplit(baseuri)
        self._scheme  = parseduri.scheme
        self._host    = parseduri.netloc
        self._path    = parseduri.path
        self._http2   = httplib2.Http()
        return

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()
        return

    def close(self):
        self._key   = None
        self._http2 = None
        return

    def baseuri(self):
        return self._baseuri

    def getpathuri(self, uripath):
        # str used here so rdflib.URIRef values can be accepted
        return urlparse.urljoin(self._baseuri, str(uripath))

    def error(self, msg, value=None):
        return HTTP_Error(msg=msg, value=value, uri=self._baseuri)

    def parseLinks(self, headers):
        """
        Parse link header(s), return dictionary of links keyed by link relation type
        """
        return parseLinks(headers["_headerlist"])

    def doRequest(self, uripath,
            method="GET", body=None, ctype=None, accept=None, reqheaders=None, exthost=False):
        """
        Perform HTTP request.

        Parameters:
            uripath     URI reference of resource to access, resolved against the base URI of
                        the current HTTP_Session object.
            method      HTTP method to use (default GET)
            body        request body to use (default none)
            ctype       content-type of request body (default none)
            accept      string containing list of content types for HTTP accept header
            reqheaders  dictionary of additional header fields to send with the HTTP request
            exthost     True if a request to a URI with a scheme and/or host different than 
                        the session base URI is to be respected (default False).

        Return:
             status, reason(text), response headers, response body

        Note: relies on to,meout or object deallocation to close connections:
        see http://stackoverflow.com/questions/16687033/is-this-a-bug-of-httplib2.
        Python's reference counting should ensure Http objects amnd referenced connections
        are deallocated promptly.
        """
        # Construct request path
        urifull  = self.getpathuri(uripath)
        uriparts = urlparse.urlsplit(urifull)
        path     = uriparts.path
        if uriparts.query: path += ("?"+uriparts.query)
        # Sort out HTTP connection to use: session or new
        if ( (uriparts.scheme and uriparts.scheme != self._scheme) or
             (uriparts.netloc and uriparts.netloc != self._host) ):
            if exthost:
                usescheme  = uriparts.scheme
                usekey     = None
            elif (uriparts.scheme and uriparts.scheme != self._scheme):
                raise HTTP_Error(
                    "URI scheme mismatch",
                    value=uriparts.scheme,
                    uri=self._baseuri)
            elif (uriparts.netloc and uriparts.netloc != self._host):
                raise HTTP_Error(
                    "URI host:port mismatch",
                    value=uriparts.netloc,
                    uri=self._baseuri)
        else:
            usescheme  = self._scheme
            usekey     = self._key
        # Assemble request headers
        if not reqheaders:
            reqheaders = {}
        if usekey:
            reqheaders["authorization"] = "Bearer "+usekey
        if ctype:
            reqheaders["content-type"] = ctype
        if accept:
            reqheaders["accept"] = accept
        # Execute request
        log.debug("HTTP_Session.doRequest method:     "+method)
        log.debug("HTTP_Session.doRequest path:       "+path)
        log.debug("HTTP_Session.doRequest reqheaders: "+repr(reqheaders))
        log.debug("HTTP_Session.doRequest body:       "+repr(body))
        (resp, data) = self._http2.request(urifull, 
            method=method, body=body, headers=reqheaders)
        # Pick out elements of response
        try:
            status   = resp.status
            reason   = resp.reason
            headerlist = [ (hn.lower(),resp[hn]) for hn in resp ]
            headers  = dict(headerlist)   # dict(...) keeps last result of multiple keys
            headers["_headerlist"] = headerlist
            log.debug("HTTP_Session.doRequest response:   "+str(status)+" "+reason)
            log.debug("HTTP_Session.doRequest rspheaders: "+repr(headers))
        except Exception, e:
            log.warn("HTTP_Session error %r accessing %s with request headers %r"%(e, uripath, reqheaders))
            status = 900
            reason = str(e)
            headers = {"_headerlist": []}
            data = None
        ###log.debug("HTTP_Session.doRequest data:     "+repr(data))
        return (status, reason, headers, data)

    def doRequestFollowRedirect(self, uripath, 
            method="GET", body=None, ctype=None, accept=None, reqheaders=None, exthost=False):
        """
        Perform HTTP request, following any redirect returned.

        Parameters:
            uripath     URI reference of resource to access, resolved against the base URI of
                        the current HTTP_Session object.
            method      HTTP method to use (default GET)
            body        request body to use (default none)
            ctype       content-type of request body (default none)
            accept      string containing list of content types for HTTP accept header
            reqheaders  dictionary of additional header fields to send with the HTTP request
            exthost     True if a request to a URI with a scheme and/or host different than 
                        the session base URI is to be respected (default False).

        Return:
             status, reason(text), response headers, final URI, response body

        """
        (status, reason, headers, data) = self.doRequest(uripath,
            method=method, accept=accept,
            body=body, ctype=ctype, reqheaders=reqheaders, 
            exthost=exthost)
        return (status, reason, headers, headers['content-location'], data)

# End.



"""
Display Annalist server help messages.
"""

from __future__ import print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import sys
import logging
import subprocess

log = logging.getLogger(__name__)

import am_errors

command_summary_help = ("\n"+
    "Commands:\n"+
    "\n"+
    "  %(prog)s help [command]\n"+
    "  %(prog)s runtests\n"+
    "  %(prog)s initialize [ CONFIG ]\n"+
    # "  %(prog)s idprovider ...\n"+  #@@ TODO
    "  %(prog)s createadminuser [ username [ email [ firstname [ lastname ] ] ] ] [ CONFIG ]\n"+
    "  %(prog)s defaultadminuser [ CONFIG ]\n"+
    "  %(prog)s updateadminuser [ username ] [ CONFIG ]\n"+
    "  %(prog)s setdefaultpermissions [ permissions ] [ CONFIG ]\n"+
    "  %(prog)s setpublicpermissions [ permissions ] [ CONFIG ]\n"+
    "  %(prog)s deleteuser [ username ] [ CONFIG ]\n"+
    "  %(prog)s createsitedata [ CONFIG ]\n"+
    "  %(prog)s updatesitedata [ CONFIG ]\n"+
    "  %(prog)s runserver [ CONFIG ]\n"+
    "  %(prog)s sitedirectory [ CONFIG ]\n"+
    "  %(prog)s serverlog [ CONFIG ]\n"+
    "  %(prog)s version\n"+
    "")

config_options_help = (
    "Annalist can be run in a number of configurations, notably\n"+
    "'development', 'personal' and 'shared'.\n"+
    "\n"+
    "A configuration can be selected by using one of the following options:\n"+
    "--devel    selects the 'development' configuration, which stores all site data\n"+
    "           within the source code tree, and configuration data in the user's\n"+
    "           home directory ('~/.annalist/')\n"+
    "--personal selects the 'personal' configuration, which stores all site data\n"+
    "           and configuration data in the activating user's home directory\n"+
    "           ('~/annalist_site/' and '~/.annalist/')\n"+
    "--shared   selects the 'shared' configuration, which stores all site and configuration\n"+
    "           data in system directories '/var/annalist_site', and configuration.\n"+
    "           data in '/etc/annalist/'\n"+
    "--configuration=NAME\n"+
    "           allows selection of any named configuration, where configuration files\n"+
    "           are stored in the Annalist source tree as '.../annalist_site/settings/NAME.py'\n"+
    "\n"+
    "The above options may be abbreviated as '-d', '-p', '-s' and '-c' respectively.\n"+
    "If no configuration is explicitly specified, '--personal' is used.\n"+
    "")

permissions_help = (
    "The 'permissions' parameter is a list of space-separated permission keywords,\n"+
    "or may be empty.  If multiple permissions are specified, some form of command-line\n"+
    "quoting should be used so they are presented as a single argument (e.g. enclose\n"+
    "the list of keywords in double quoted).\n"+
    "If not specified on the command line, the user will be prompted for default permissions.\n"+
    "\n"+
    "Initially defined permissions are:\n"+
    "CREATE_COLLECTION   site-level permission required to create new collection (or ADMIN).\n"+
    "DELETE_COLLECTION   site-level permission required to delete a collection (or ADMIN).\n"+
    "VIEW                permission to view or list data in a collection\n"+
    "CREATE              permission to create new data in a collection\n"+
    "UPDATE              permission to update existing data in a collection\n"+
    "DELETE              permission to delete data from a collection\n"+
    "CONFIG              permission to add or modify configuration data for a collection\n"+
    "                    (i.e. types, views, lists, fields, and field groups)\n"+
    "ADMIN               permission to add or modify user permissions\n"+
    "")

def am_help(options, progname):
    """
    Display annalist-manager command help

    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    if len(options.args) > 1:
        print("Unexpected arguments for %s: (%s)"%(options.command, " ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    status = am_errors.AM_SUCCESS
    if len(options.args) == 0:
        help_text = (
            command_summary_help+
            "\n"+
            "For more information about command options, use:\n"+
            "\n"+
            "  %(prog)s --help\n"+
            "")
    elif options.args[0].startswith("runt"):
        help_text = ("\n"+
            "  %(prog)s runtests\n"+
            "\n"+
            "Runs annalist test suite using installed software\n"+
            "\n"+
            "")
    elif options.args[0].startswith("init"):
        help_text = ("\n"+
            "  %(prog)s initialize [ CONFIG ]\n"+
            "\n"+
            "Initializes the installed software for an indicated configuration.\n"+
            "Mainly, this involves creating the internal database used to manage users, etc.\n"+
            "\n"+
            config_options_help+
            "\n"+
            "")
    elif options.args[0].startswith("createa"):
        help_text = ("\n"+
            "  %(prog)s createadminuser [ CONFIG ]\n"+
            "\n"+
            "Creates an Annalist administrative user.\n"+
            "\n"+
            "The software will prompt for a username, email address and password.\n"+
            "\n"+
            "The administrative user details can be used to log in to Annalist using\n"+
            "the 'Local user credentials: login' link at the bottom of the login page.\n"+
            "An administrative user can then use the 'Admin' link at the bottom of other\n"+
            "Annalist pages to create, modify or delete other local user credentials.\n"+
            "\n"+
            "Annalist is intended to be used with a federated authentication service,\n"+
            "such as Google+, but setting up such a service can be tricky, and for evaluation\n"+
            "or personal-only use it may be quicker to use locally managed user credentials\n"+
            "\n"+
            config_options_help+
            "\n"+
            "")
    elif options.args[0].startswith("defaulta"):
        help_text = ("\n"+
            "  %(prog)s defaultadminuser [ CONFIG ]\n"+
            "\n"+
            "Creates a default Annalist administrative user.\n"+
            "\n"+
            "The software uses default values for a username and email address, prompts for\n"+
            "a password, and creates a new admin user with username 'admin'.\n"+
            "\n"+
            "The administrative user details can be used to log in to Annalist using\n"+
            "the 'Local user credentials: login' link at the bottom of the login page.\n"+
            "An administrative user can then use the 'Admin' link at the bottom of other\n"+
            "Annalist pages to create, modify or delete other local user credentials.\n"+
            "\n"+
            "Annalist is intended to be used with a federated authentication service,\n"+
            "such as Google+, but setting up such a service can be tricky, and for evaluation\n"+
            "or personal-only use it may be quicker to use locally managed user credentials\n"+
            "\n"+
            config_options_help+
            "\n"+
            "")
    elif options.args[0].startswith("updatea"):
        help_text = ("\n"+
            "  %(prog)s updateadminuser [ username ] [ CONFIG ]\n"+
            "\n"+
            "Updates an existing Django user to admin status; i.e. they are assigned 'staff'\n"+
            "and 'superuser' attributes in the Django user database, and assigned site-wide\n"+
            "ADMIN permissions in the Annalist site indicated by CONFIG.\n"+
            "\n"+
            config_options_help+
            "\n"+
            "")
    elif options.args[0].startswith("setdef"):
        help_text = ("\n"+
            "  %(prog)s setdefaultpermissions [ permissions ] [ CONFIG ]\n"+
            "\n"+
            "Sets site-wide default permissions for logged-in users in the\n"+
            "Annalist site indicated by CONFIG.  These permissions are superseded by\n"+
            "any permissions defined specifically for a logged-in user, or by\n"+
            "user '_default_user_perms' entry defined for any collection.\n"+
            "\n"+
            permissions_help+
            "\n"+
            config_options_help+
            "\n"+
            "")
    elif options.args[0].startswith("setpub"):
        help_text = ("\n"+
            "  %(prog)s setpublicpermissions [ permissions ] [ CONFIG ]\n"+
            "\n"+
            "Sets site-wide public access permissions (i.e. for requests where there is no active login)\n"+
            "in the Annalist site indicated by CONFIG.  These permissions may be superseded by\n"+
            "'_unknown_user_perms' permissions defined for any specific collection.\n"+
            "\n"+
            permissions_help+
            "\n"+
            config_options_help+
            "\n"+
            "")
    elif options.args[0].startswith("deleteu"):
        help_text = ("\n"+
            "  %(prog)s deleteuser [ username ] [ CONFIG ]\n"+
            "\n"+
            "Deletes the specified Django user, and also removes any site-wide permissions defined\n"+
            "for that user in he Annalist site indicated by CONFIG.\n"+
            "\n"+
            config_options_help+
            "\n"+
            "")
    elif options.args[0].startswith("creates"):
        help_text = ("\n"+
            "  %(prog)s createsite [ CONFIG ] [ --force | -f ]\n"+
            "\n"+
            "Creates Annalist empty site data.\n"+
            "\n"+
            "Creates empty site data (i.e. with no collections) for an Annalist service.\n"+
            "\n"+
            "If the site already exists, the command is refused unless the '--force' or '-f'\n"+
            "option is given.\n"+
            "\n"+
            config_options_help+
            "\n"+
            "")
    elif options.args[0].startswith("updates"):
        help_text = ("\n"+
            "  %(prog)s updatesite [ CONFIG ]\n"+
            "\n"+
            "Updates the site-wide data in an existing annalist site.\n"+
            "\n"+
            "\nExisting collection data is left untouched.\n"+
            "\n"+
            "If the site does not exist, the command fails.\n"+
            "\n"+
            config_options_help+
            "\n"+
            "")
    elif options.args[0].startswith("runs"):
        help_text = ("\n"+
            "  %(prog)s runserver [ CONFIG ]\n"+
            "\n"+
            "Starts an Annalist server running.\n"+
            "\n"+
            config_options_help+
            "\n"+
            "")
    elif options.args[0].startswith("site"):
        help_text = ("\n"+
            "  %(prog)s sitedirectory [ CONFIG ]\n"+
            "\n"+
            "Sends the name of Annalist site directory to standard output.\n"+
            "\n"+
            "This is a convenience function to locate the site data directory, which\n"+
            "may be buried deep in the Python virtual environment files.\n"+
            "\n"+
            config_options_help+
            "\n"+
            "")
    elif options.args[0].startswith("ver"):
        help_text = ("\n"+
            "  %(prog)s version\n"+
            "\n"+
            "Sends the Annalist software version string to standard output.\n"+
            "\n"+
            "")
    else:
        help_text = "Unrecognized command for %s: (%s)"%(options.command, options.args[0])
        status = am_errors.AM_UNKNOWNCMD
    print(help_text%{'prog': progname}, file=sys.stderr)
    return status

# End.



"""
Annalist collection

A collection is represented by:
- an ID (slug)
- a URI
- a name/label
- a description
- a set of record types
- a set of list views
- a set of record views
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import os.path
import urlparse
import shutil

import logging
log = logging.getLogger(__name__)

from django.conf import settings

from annalist                       import layout
from annalist.exceptions            import Annalist_Error
from annalist.identifiers           import RDF, RDFS, ANNAL
from annalist                       import util

from annalist.models.entity         import Entity
from annalist.models.annalistuser   import AnnalistUser
from annalist.models.recordtype     import RecordType
from annalist.models.recordview     import RecordView
from annalist.models.recordlist     import RecordList

class Collection(Entity):

    _entitytype     = ANNAL.CURIE.Collection
    _entitytypeid   = "_collection"
    _entityview     = layout.SITE_COLL_VIEW
    _entitypath     = layout.SITE_COLL_PATH
    _entityfile     = layout.COLL_META_FILE
    _entityref      = layout.META_COLL_REF

    def __init__(self, parentsite, coll_id):
        """
        Initialize a new Collection object, without metadta (yet).

        parentsite  is the parent site from which the new collection is descended.
        coll_id     the collection identifier for the collection
        """
        super(Collection, self).__init__(parentsite, coll_id)
        self._parentsite = parentsite
        self._types_by_id  = None
        self._types_by_uri = None
        return

    # Site

    def get_site(self):
        """
        Return site object for the site from which the current collection is accessed.
        """
        return self._parentsite

    # User permissions

    def create_user_permissions(self, user_id, user_uri,
            user_name, user_description,
            user_permissions=["VIEW"]
            ):
        user_values = (
            { ANNAL.CURIE.type:             ANNAL.CURIE.User
            , RDFS.CURIE.label:             user_name
            , RDFS.CURIE.comment:           user_description
            , ANNAL.CURIE.user_uri:         user_uri
            , ANNAL.CURIE.user_permissions: user_permissions
            })
        user = AnnalistUser.create(self, user_id, user_values)
        return user

    def get_user_permissions(self, user_id, user_uri):
        """
        Get a user permissions record (AnnalistUser).

        To return a value, both the user_id and the user_uri (typically a mailto: URI, but
        may be any *authenticated* identifier) must match.  This is to prevent access to 
        records of a deleted account being granted to a new account created with the 
        same user_id (username).

        user_id         local identifier for the type to retrieve.
        user_uri        authenticated identifier associated with the user_id.  That is,
                        the authentication service used is presumed to confirm that
                        the identifier belongs to the user currently logged in with
                        the supplied username.

        returns an AnnalistUser object for the identified user, or None.  This object contains
                information about permissions granted to the user in the current collection.
        """
        user = AnnalistUser.load(self, user_id, altparent=self._parentsite)
        log.debug("Collection.get_user_permissions: user_id %s, user_uri %s, user %r"%
            (user_id, user_uri, user)
            )
        if user:
            for f in [RDFS.CURIE.label, RDFS.CURIE.comment, ANNAL.CURIE.user_uri, ANNAL.CURIE.user_permissions]:
                if f not in user:
                    user = None
                    break
        if user and user[ANNAL.CURIE.user_uri] != user_uri:
            user = None         # URI mismatch: return None.
        return user

    # Record types

    def _update_type_cache(self, type_entity):
        """
        Add single type entity to type cache
        """
        if type_entity:
            self._types_by_id[type_entity.get_id()]   = type_entity
            self._types_by_uri[type_entity.get_uri()] = type_entity
        return

    def _flush_type(self, type_id):
        """
        Remove single identified type entity from type cache
        """
        if self._types_by_id:
            t = self._types_by_id.get(type_id, None)
            if t:
                type_uri = t.get_uri()
                self._types_by_id.pop(type_id, None)
                self._types_by_uri.pop(type_uri, None)
        return

    def _load_types(self):
        """
        Initialize cache of RecordType entities
        """
        if not (self._types_by_id and self._types_by_uri):
            self._types_by_id  = {}
            self._types_by_uri = {}
            for type_id in self._children(RecordType, altparent=self._parentsite):
                t = RecordType.load(self, type_id, altparent=self._parentsite)
                self._update_type_cache(t)
        return

    def types(self, include_alt=True):
        """
        Generator enumerates and returns record types that may be stored
        """
        altparent = self._parentsite if include_alt else None
        for f in self._children(RecordType, altparent=altparent):
            t = self.get_type(f)
            if t and t.get_id() != "_initial_values":
                yield t
        return

    def add_type(self, type_id, type_meta):
        """
        Add a new record type to the current collection

        type_id     identifier for the new type, as a string
                    with a form that is valid as URI path segment.
        type_meta   a dictionary providing additional information about
                    the type to be created.

        Returns a RecordType object for the newly created type.
        """
        t = RecordType.create(self, type_id, type_meta)
        if self._types_by_id:
            self._update_type_cache(t)
        return t

    def get_type(self, type_id):
        """
        Retrieve identified type description

        type_id     local identifier for the type to retrieve.

        returns a RecordType object for the identified type, or None.
        """
        self._load_types()
        t = self._types_by_id.get(type_id, None)
        # Was it created but not cached?
        if not t and RecordType.exists(self, type_id, altparent=self._parentsite):
            t = RecordType.load(self, type_id, altparent=self._parentsite)
            self._update_type_cache(t)
        return t
        #@@
        # t = RecordType.load(self, type_id, altparent=self._parentsite)
        #@@

    def get_uri_type(self, type_uri, include_alt=True):
        """
        Return type entity corresponding to the supplied type URI
        """
        self._load_types()
        t = self._types_by_uri.get(type_uri, None)
        return t
        #@@
        # for t in self.types(include_alt=include_alt):
        #     turi = t.get_uri()
        #     if turi == type_uri:
        #         return t
        #@@

    def remove_type(self, type_id):
        """
        Remove identified type description

        type_id     local identifier for the type to remove.

        Returns a non-False status code if the type is not removed.
        """
        self._flush_type(type_id)
        s = RecordType.remove(self, type_id)
        return s

    # Record views

    def views(self, include_alt=True):
        """
        Generator enumerates and returns record views that may be stored
        """
        altparent = self._parentsite if include_alt else None
        for f in self._children(RecordView, altparent=altparent):
            v = self.get_view(f)
            if v and v.get_id() != "_initial_values":
                yield v
        return

    def add_view(self, view_id, view_meta):
        """
        Add a new record view to the current collection

        view_id     identifier for the new view, as a string
                    with a form that is valid as URI path segment.
        view_meta   a dictionary providing additional information about
                    the view to be created.

        returns a RecordView object for the newly created view.
        """
        v = RecordView.create(self, view_id, view_meta)
        return v

    def get_view(self, view_id):
        """
        Retrieve identified view description

        view_id     local identifier for the view to retrieve.

        returns a RecordView object for the identified view, or None.
        """
        v = RecordView.load(self, view_id, altparent=self._parentsite)
        return v

    def remove_view(self, view_id):
        """
        Remove identified view description

        view_id     local identifier for the view to remove.

        Returns a non-False status code if the view is not removed.
        """
        s = RecordView.remove(self, view_id)
        return s

    # Record lists

    def lists(self, include_alt=True):
        """
        Generator enumerates and returns record lists that may be stored
        """
        altparent = self._parentsite if include_alt else None
        for f in self._children(RecordList, altparent=altparent):
            l = self.get_list(f)
            if l and l.get_id() != "_initial_values":
                yield l
        return

    def add_list(self, list_id, list_meta):
        """
        Add a new record list to the current collection

        list_id     identifier for the new list, as a string
                    with a form that is valid as URI path segment.
        list_meta   a dictionary providing additional information about
                    the list to be created.

        returns a RecordList object for the newly created list.
        """
        l = RecordList.create(self, list_id, list_meta)
        return l

    def get_list(self, list_id):
        """
        Retrieve identified list description

        list_id     local identifier for the list to retrieve.

        returns a RecordList object for the identified list, or None.
        """
        l = RecordList.load(self, list_id, altparent=self._parentsite)
        return l

    def remove_list(self, list_id):
        """
        Remove identified list description

        list_id     local identifier for the list to remove.

        Returns a non-False status code if the list is not removed.
        """
        s = RecordList.remove(self, list_id)
        return s

    def set_default_list(self, list_id):
        """
        Set and save the default list to be displayed for the current collection.
        """
        self[ANNAL.CURIE.default_list] = list_id
        self._save()
        return

    def get_default_list(self):
        """
        Return the default list to be displayed for the current collection.
        """
        list_id = self.get(ANNAL.CURIE.default_list, None)
        if list_id and not RecordList.exists(self, list_id, altparent=self._parentsite):
            log.warning(
                "Default list %s for collection %s does not exist"%
                (list_id, self.get_id())
                )
            list_id = None
        return list_id 
# End.



"""
OAuth2 / OpenID Connect authentication related views
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

# @@TODO: refactor all OAuth2 details from views to this module
# @@TODO: define a view decorator to apply OAuth2 authentication requirement

import os
import re
import json
import copy
import uuid
import urllib
from urlparse import urlparse, urljoin
from importlib import import_module

import logging
log = logging.getLogger(__name__)

import httplib2

from oauth2client.client import OAuth2WebServerFlow
from oauth2client import xsrfutil
from oauth2client.client import flow_from_clientsecrets, FlowExchangeError
from oauth2client.django_orm import Storage

from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.template import RequestContext, loader
from django.views import generic
from django.views.decorators.csrf import csrf_exempt

from django.contrib.auth import authenticate, login, logout
from django.contrib.auth.models import User

settings = import_module(os.environ["DJANGO_SETTINGS_MODULE"])

from utils.http_errors import error400values

from models import CredentialsModel

# Per-instance generated secret key for CSRF protection via OAuth2 state value.
# Regenerated each time this service is started.
FLOW_SECRET_KEY = str(uuid.uuid1())

PROVIDER_LIST = None

CLIENT_SECRETS = None

SCOPE_DEFAULT = "openid profile email"

OAuth2WebServerFlow_strip = (
    "step1_get_authorize_url",
    "step2_exchange"
    )

def collect_client_secrets():
    global CLIENT_SECRETS, PROVIDER_LIST
    if CLIENT_SECRETS is None:
        CLIENT_SECRETS = {}
        PROVIDER_LIST  = {}
        clientsecrets_dirname = os.path.join(settings.CONFIG_BASE, "providers/")
        if os.path.isdir(clientsecrets_dirname):
            clientsecrets_files   = os.listdir(clientsecrets_dirname)
            for f in clientsecrets_files:
                p = os.path.join(clientsecrets_dirname,f)
                j = json.load(open(p, "r"))
                n = j['web']['provider']
                CLIENT_SECRETS[n] = j['web']
                PROVIDER_LIST[n]  = p
    return

def object_to_dict(obj, strip):
    """Utility function that creates dictionary representation of an object.

    Args:
        strip: an array of names of members to not include in the dict.

    Returns:
        dictionary, with non-excluded values that can be used to reconstruct an instance
        of the object via its constructor (assuming an appropriate constructor form, as
        used below for dict_to_flow)
    """
    t = type(obj)
    d = copy.copy(obj.__dict__)
    for member in strip:
      if member in d:
        del d[member]
    d['_class'] = t.__name__
    d['_module'] = t.__module__
    return d

def flow_to_dict(f):
    return object_to_dict(f, OAuth2WebServerFlow_strip) 

def dict_to_flow(d):
    """
    Constructs a OAuth2WebServerFlow object from a dictionary previously created
    by flow_to_dict.

    Args:
        d:  dict, generated by object_to_dict
    """
    flow = OAuth2WebServerFlow(
        d['client_id'], d['client_secret'], d['scope'],
        redirect_uri=d['redirect_uri'], 
        user_agent=d['user_agent'],
        auth_uri=d['auth_uri'], 
        token_uri=d['token_uri'],
        revoke_uri=d['revoke_uri'],
        **d['params']
        )
    return flow

def authentication_required(
        login_form_uri=None, login_post_uri=None, login_done_uri=None, 
        continuation_url=None, scope=SCOPE_DEFAULT):
    """
    Decorator for view handler function that activates OAuth2 authentication flow
    if the current request is not already associated with an authenticated user.
    """
    # @@NOTE: not tested; the mix of static and dynamic parameters required makes
    #         the in-line form easier to use than a decorator.
    def decorator(func):
        def guard(view, values):
            return (
                confirm_authentication(view, 
                    login_form_uri, login_post_uri, login_done_uri, 
                    continuation_url, scope)
            or
                func(view, values)
            )
        return guard
    return decorator

def HttpResponseRedirectWithQuery(redirect_uri, query_params):
    nq = "?"
    for pname in query_params.keys():
        redirect_uri += nq + pname + "=" + urllib.quote(query_params[pname])
        nq = "&"
    # log.info("redirect_uri: "+redirect_uri)
    return HttpResponseRedirect(redirect_uri)

def HttpResponseRedirectLoginWithMessage(request, message):
    login_form_uri = request.session['login_form_uri']
    log.info("login_form_uri: "+login_form_uri)
    query_params = (
        { "continuation": request.session['continuation_url']
        , "scope":        request.session['oauth2_scope']
        , "message":      message
        })
    return HttpResponseRedirectWithQuery(login_form_uri, query_params)

# Authentication and authorization
def confirm_authentication(view, 
        login_form_uri=None, login_post_uri=None, login_done_uri=None, 
        continuation_url=None, scope=SCOPE_DEFAULT):
    """
    Return None if required authentication is present, otherwise
    a login redirection response to the supplied URI

    view.credential is set to credential that can be used to access resource
    """
    if view.request.user.is_authenticated():
        storage         = Storage(CredentialsModel, 'id', view.request.user, 'credential')
        view.credential = storage.get()
        # log.info("view.credential %r"%(view.credential,))
        if view.credential is not None:
            if not view.credential.invalid:
                return None         # Valid credential present: proceed...
        else:
            # Django login with local credential: check for user email address
            #
            # @@TODO: is this safe?
            # 
            # NOTE: currently, view.credential is provided by the oauth2 and used
            # only for the .invalid test above.  If it is ever used by other 
            # application components, it may be necessary to construct a
            # credential for local logins.  In the long run, if credentials will
            # be used to access third party services or resources, it may not be 
            # possible to use non-Oauth2 credentials here.  In the meanwhile,
            # allowing local Django user credentials provides an easier route for
            # getting a software instance installed for evaluation purposes.
            #
            if view.request.user.email:
                return None        # Assume valid login: proceed...
            else:
                return error400values(view, "Local user has no email address")
    if not login_form_uri:
        return error400values(view, "No login form URI specified")
    if not login_done_uri:
        return error400values(view, "No login completion URI specified")
    if not login_post_uri:
        login_post_uri = login_form_uri
    if not continuation_url:
        continuation_url = view.request.path
    # Redirect to initiate login sequence 
    view.request.session['login_form_uri']   = login_form_uri
    view.request.session['login_post_uri']   = login_post_uri
    view.request.session['login_done_uri']   = login_done_uri
    view.request.session['continuation_url'] = continuation_url
    view.request.session['oauth2_scope']     = scope
    query_params = (
        { "continuation": continuation_url
        , "scope":        scope
        , "message":      ""
        })
    return HttpResponseRedirectWithQuery(login_form_uri, query_params)

class LoginUserView(generic.View):
    """
    View class to present login form to gather user id and other login information.

    The login page solicits a user id and an identity provider

    The login page supports the following request parameters:

    continuation={uri}
    - a URI that is retrieved, with a suitable authorization grant as a parameter, 
      when appropriate permission has been confirmed by an authenticated user.
    scope={string}
    - requested or required access scope
    """

    def get(self, request):
        collect_client_secrets()
        # @@TODO: check PROVIDER_LIST, report error if none here
        # Retrieve request parameters
        continuation = request.GET.get("continuation", "/no-login-continuation/")
        scope        = request.GET.get("scope",        SCOPE_DEFAULT)
        message      = request.GET.get("message",      "")
        # Check required values in session - if missing, restart sequence from original URI
        # This is intended to avoid problems if this view is invoked out of sequence
        login_post_uri = request.session.get('login_post_uri', None)
        login_done_uri = request.session.get('login_done_uri', None)
        if (login_post_uri is None) or (login_done_uri is None):
            return HttpResponseRedirect(continuation)
        # Display login form
        default_provider = ""
        if len(PROVIDER_LIST.keys()) > 0:
            default_provider = PROVIDER_LIST.keys()[0]
        logindata = (
            { "login_post":     request.session['login_post_uri']
            , "login_done":     request.session['login_done_uri']
            , "continuation":   continuation
            , "userid":         request.GET.get("userid", "")
            , "providers":      PROVIDER_LIST.keys()
            , "scope":          scope
            , "message":        message
            , "suppress_user":  True
            # Default provider
            , "provider":       default_provider
            , 'help_filename':  'login-help'
            })
        # Load help text if available
        # @@NOTE: this next is dependent on data provided by the calling app?
        if 'help_filename' in logindata:
            help_filepath = os.path.join(
                settings.SITE_SRC_ROOT, 
                "annalist/views/help/%s.html"%logindata['help_filename']
                )
            if os.path.isfile(help_filepath):
                with open(help_filepath, "r") as helpfile:
                    logindata['help_text'] = helpfile.read()
        # Render form & return control to browser
        template = loader.get_template('login.html')
        context  = RequestContext(self.request, logindata)
        return HttpResponse(template.render(context))

class LoginPostView(generic.View):
    """
    View class initiate an OAuth2 authorization (or similar) flow, typically on POST
    of a login form.

    It saves the supplied user id in a session value, and redirects the user to the 
    identity provider, which in due course returns control to the application along 
    with a suitable authorization grant.

    The login form provides the following values:

    userid={string}
    - a user identifying string that will be associated with the external service
      login credentials.
    provider={string}
    - a string that identifiues a provioder selectred to proviode authentication/
      authorization for the indicated user.  This string is an index to PROVIDER_LIST,
      which in turn contains filenames for client secrets to user when using the 
      indicated identity provider.
    login_done={uri}
    - a URI that is retrieved, with a suitable authorization grant as a parameter, 
      when appropriate permission has been confirmed by an authenticated user.
      Communicated via a hidden form value.
    continuation={uri}
    - a URI that is retrieved, with a suitable authorization grant as a parameter, 
      when appropriate permission has been confirmed by an authenticated user.
      Communicated via a hidden form value.
    scope={string}
    - Requested or required access scope, communicated via a hidden form value.
    """

    def post(self, request):
        # Retrieve request parameters
        userid        = request.POST.get("userid",        "")
        provider      = request.POST.get("provider",      "Google")
        login_done    = request.POST.get("login_done",    "/no_login_done_in_form_response/")
        continuation  = request.POST.get("continuation",  "/no_continuation_in_form_response/")
        scope         = request.POST.get("scope",         SCOPE_DEFAULT) 
        # Access or create flow object for this session
        if request.POST.get("login", None) == "Login":
            collect_client_secrets()
            # Create and initialize flow object
            clientsecrets_filename = os.path.join(
                settings.CONFIG_BASE, "providers/", PROVIDER_LIST[provider]
                )
            flow = flow_from_clientsecrets(
                clientsecrets_filename,
                scope=scope,
                redirect_uri=request.build_absolute_uri(login_done)
                )
            flow.params['state']        = xsrfutil.generate_token(FLOW_SECRET_KEY, request.user)
            flow.params['provider']     = provider
            flow.params['userid']       = userid
            # flow.params['scope']        = scope
            flow.params['continuation'] = continuation
            # Save flow object in Django session
            request.session['oauth2flow'] = flow_to_dict(flow)
            # Initiate OAuth2 dance
            auth_uri = flow.step1_get_authorize_url()
            return HttpResponseRedirect(auth_uri)
        # Login cancelled: redirect to continuation
        # (which may just redisplay the login page)
        return HttpResponseRedirect(continuation)

class LoginDoneView(generic.View):
    """
    View class used to complete login process with authorization grant provided by
    authorization server.
    """

    def get(self, request):
        # Look for authorization grant
        flow   = dict_to_flow(request.session['oauth2flow'])
        userid = flow.params['userid']
        if not userid:
            log.info("No User ID specified")
            return HttpResponseRedirectLoginWithMessage(request, "No User ID specified")
        if not re.match(r"\w+$", userid):
            return HttpResponseRedirectLoginWithMessage(request, 
                "User ID must consist of letters, digits and '_' chacacters (%s)"%(userid))
        # Save copy of current user details, if defined
        try:
            olduser = User.objects.get(username=userid)
        except User.DoesNotExist:
            olduser = None
        # Get authenticated user details
        try:
            credential = flow.step2_exchange(request.REQUEST) # Raises FlowExchangeError if a problem occurs
            authuser = authenticate(
                username=userid, password=credential, 
                profile_uri=CLIENT_SECRETS[flow.params['provider']]['profile_uri']
                )
        except FlowExchangeError, e:
            log.error("CLIENT_SECRETS %r"%(CLIENT_SECRETS[flow.params['provider']],))
            return HttpResponseRedirectLoginWithMessage(request, str(e))
        # Check authenticated details for user id match any previous values.
        #
        # The user id is entered by the user on the login form, and is used as a key to
        # access authenticated user details in the Django user database.  The user id 
        # itself is not checked by the Oauth2 login flow, other than for checking that
        # it containbs only work characters
        #
        # Instead, we trust that the associated email address has been confirmed by the 
        # OAuth2 provider, and don't allow login where the email adress differs from any 
        # currently saved email address for the user id used..  This aims to  prevent a 
        # new set of OAuth2 credentials being used for a previously created Django user id.
        #
        if not authuser.email:
            return HttpResponseRedirectLoginWithMessage(request, 
                "No email address associated with authenticated user %s"%(userid))
        if olduser:
            if authuser.email != olduser.email:
                return HttpResponseRedirectLoginWithMessage(request, 
                    "Authenticated user %s email address mismatch (%s, %s)"%
                        (userid, authuser.email, olduser.email))
        # Complete the login and save details
        authuser.save()
        login(request, authuser)
        storage    = Storage(CredentialsModel, 'id', request.user, 'credential')
        storage.put(credential)
        # Don't normally log the credential/token as they might represent a security leakage:
        # log.debug("LoginDoneView: credential:      "+repr(credential.to_json()))
        # log.info("LoginDoneView: id_token:        "+repr(credential.id_token))
        log.info("LoginDoneView: user.username:   "+authuser.username)
        log.info("LoginDoneView: user.first_name: "+authuser.first_name)
        log.info("LoginDoneView: user.last_name:  "+authuser.last_name)
        log.info("LoginDoneView: user.email:      "+authuser.email)
        return HttpResponseRedirect(flow.params['continuation'])

class LogoutUserView(generic.View):
    """
    View class to handle logout
    """

    def get(self, request):
        logout(request)
        continuation = request.GET.get("continuation", urljoin(urlparse(request.path).path, "../"))
        return HttpResponseRedirect(continuation)

# End.



"""
Annalist model admin hooks
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

from django.contrib import admin

# Register your models here.



"""
This module contains utilities for use in conjunction with field renderers.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import traceback
import re

from urlparse               import urljoin  # py3: from urllib.parse ...
from collections            import OrderedDict, namedtuple

from django.conf            import settings

from annalist.exceptions    import TargetIdNotFound_Error, TargetEntityNotFound_Error
from annalist.identifiers   import RDFS, ANNAL
from annalist.util          import split_type_entity_id

from annalist.models.entitytypeinfo         import EntityTypeInfo
from annalist.models.entity                 import EntityRoot

from annalist.views.uri_builder             import (
    uri_params, uri_with_params, continuation_params
    )
from annalist.views.form_utils.fieldchoice  import FieldChoice


class bound_field(object):
    """
    Class representing an entity bound to a field description, 
    which can be used as input for data driven rendering of a form field.

    The field description contains information used to extract the field
    value from the entity.

    This class, and in particular its `__getattr__` method, is provided to 
    allow indirected access to an entity value fields to be performed using a 
    Django template using, e.g., "{{ field.field_value }}" (thus satisfying the 
    Django design goal that computation is handled in the python code rather 
    than the template, though in this case the computation us handled while
    rather than before the page is rendered).

    See also: http://docs.python.org/2/reference/datamodel.html#slots

    >>> entity = EntityRoot("entityuri", "entitydir")
    >>> entity.set_id("testentity")
    >>> vals = entity.set_values({"foo": "foo_val", "bar": "bar_val"})
    >>> field_foo_desc = {"field_id": "foo_id", "field_property_uri": "foo", "field_type": "foo_type"}
    >>> field_foo = bound_field(field_foo_desc, entity)
    >>> field_foo.field_type
    'foo_type'
    >>> field_foo.field_value
    'foo_val'
    >>> field_bar_desc = {"field_id": "bar_id", "field_property_uri": "bar", "field_type": "bar_type"}
    >>> field_bar = bound_field(field_bar_desc, entity)
    >>> field_bar.field_type
    'bar_type'
    >>> field_bar.field_value
    'bar_val'
    >>> field_def_desc = {"field_id": "def_id", "field_property_uri": "def", "field_type": "def_type"}
    >>> entityvals = entity.get_values()
    >>> entityvals['entity_id']      = entity.get_id()
    >>> entityvals['entity_type_id'] = entity.get_type_id()
    >>> entityvals['entity_link']    = entity.get_url()
    >>> field_def = bound_field(field_def_desc, entity)
    >>> field_def.field_type
    'def_type'
    >>> field_def.field_placeholder in ["...","@@bound_field.field_placeholder@@"]
    True
    >>> field_def.field_value == ""
    True
    >>> field_def = bound_field(field_def_desc, entity, context_extra_values={"def": "default"})
    >>> field_def.field_type
    'def_type'
    >>> field_def.field_value
    'default'
    >>> field_def.entity_link
    'entityuri/'
    >>> field_def.htmlrepr()
    "<ul><li>key: def</li><li>val: default</li><li>field_description: {'field_property_uri': 'def', 'field_id': 'def_id', 'field_type': 'def_type'}</li></ul>"
    """

    __slots__ = ("_field_description", "_entityvals", "_targetvals", "_key", "_extras")

    def __init__(self, field_description, entityvals, context_extra_values=None):
        """
        Initialize a bound_field object.

        field_description       is a dictionary-like object describing a display
                                field.  See `FieldDescription` class for more details.
        entityvals              is an entity values dictionary from which a value to be 
                                rendered is obtained.  The specific field value used is 
                                defined by the combination with `field_description`.  
        context_extra_values    if supplied, a supplementary value dictionary that may be
                                probed for values that are not provided by the entity itself.  
                                Can be used to specify default values for an entity.
        """
        self._field_description = field_description
        self._entityvals        = entityvals
        self._targetvals        = None
        self._key               = self._field_description['field_property_uri']
        self._extras            = context_extra_values
        eid = entityvals.get('entity_id', "@@@render_utils.__init__@@@")
        # log.log(settings.TRACE_FIELD_VALUE,
        #     "bound_field: field_id %s, entity_id %s, value_key %s, value %s"%
        #     (field_description['field_id'], eid, self._key, self['field_value'])
        #     )
        return

    def __getattr__(self, name):
        """
        Get a bound field description attribute.  Broadly, if the attribute name is 
        "field_value" then the value corresponding to the field description is 
        retrieved from the entity, otherwise the named attribute is retrieved from 
        the field description.

        There are also a number of other special cases handled here as needed to 
        support internally-generated hyperlinks and internal system logic.
        """
        # log.info("self._key %s, __getattr__ %s"%(self._key, name))
        # log.info("self._key %s"%self._key)
        # log.info("self._entity %r"%self._entity)
        if name in ["entity_id", "entity_link", "entity_type_id", "entity_type_link"]:
            return self._entityvals.get(name, "")

        elif name in ["field_value", "field_edit_value"]:
            return self.get_field_value()
        elif name == "field_value_link":
            return self.get_field_link()
        elif name == "field_value_link_continuation":
            return self.get_link_continuation(self.get_field_link())

        elif name in ["target_value", "field_view_value"]:
            return self.get_target_value()
        elif name == "target_value_link":
            return self.get_target_link()
        elif name == "target_value_link_continuation":
            return self.get_link_continuation(self.get_target_link())

        elif name == "entity_link_continuation":
            return self.entity_link+self.get_continuation_param()
        elif name == "entity_type_link_continuation":
            return self.entity_type_link+self.get_continuation_param()
        elif name == "continuation_url":
            return self.get_continuation_url()
        elif name == "continuation_param":
            return self.get_continuation_param()
        elif name == "field_description":
            return self._field_description
        elif name == "field_value_key":
            return self._key
        elif name == "context_extra_values":
            return self._extras
        elif name == "options":
            return self.get_field_options()
        #@@
        # elif name == "field_placeholder":
        #     return self._field_description.get('field_placeholder', "@@bound_field.field_placeholder@@")
        #@@
        else:
            # log.info("bound_field[%s] -> %r"%(name, self._field_description[name]))
            return self._field_description.get(name, "@@bound_field.%s@@"%(name))

    def get_field_value(self):
        field_val = None
        if self._key in self._entityvals:
            field_val = self._entityvals[self._key]
        elif self._extras and self._key in self._extras:
            field_val = self._extras[self._key]
        if field_val is None:
            # Return default value, or empty string.
            # Used to populate form field value when no value supplied, or provide per-field default
            field_val = self._field_description.get('field_default_value', None)
            if field_val is None:
                field_val = ""
        return field_val

    def get_field_link(self):
        # Return link corresponding to field value that is a selection from an enumeration of entities
        # (or some other value with an associated link), or None
        #@@
        # links = self._field_description['field_choice_links']
        # v     = self.field_value
        # if links and v in links:
        #     return links[v]
        #@@
        choices = self._field_description['field_choices']  # OrderedDict
        v       = self.field_value
        if choices and v in choices:
            return choices[v].link
        return None

    def get_target_value(self):
        """
        Get target value of field for view display.

        This may be different from field_value if it references another entity field
        """
        targetvals = self.get_targetvals()
        log.debug("bound_field.get_target_value: targetvals %r"%(targetvals,))
        target_key = self._field_description.get('field_ref_field', None)
        target_key = target_key and target_key.strip()
        if targetvals is not None:
            if target_key:
                log.debug("bound_field.get_target_value: target_key %s"%(target_key,))
                target_value = targetvals.get(target_key, "(@%s)"%(target_key))
            else:
                target_value = targetvals
        elif target_key:
            target_value = self._entityvals.get(target_key, "(@%s)"%(target_key))
        else:
            target_value = self.field_value
        log.debug("bound_field.get_target_value result: %r"%(target_value,))
        return target_value

    def get_target_link(self):
        """
        Return link corresponding to target value, or None.

        The target value is treated as a relative reference relative to the field_link value.
        If the target value is itself an absolute URI, it will be used as-is.

        If the target value is a dictionary structure created by a URIImport or FileUpload field, 
        the resulting value links to the imported data object.
        """
        target_base  = self.get_field_link() or self.entity_link
        target_value = self.get_target_value()
        log.debug("get_target_link: base %r, value %r"%(target_base, target_value))
        if target_base and target_value:
            if isinstance(target_value, dict) and 'resource_name' in target_value:
                target_ref = target_value['resource_name']
            elif isinstance(target_value, (str, unicode)):
                target_ref = target_value
            else:
                log.warning(
                    "bound_field.get_target_link: "+
                    "target_value must be URI string or URIImport structure; got %r"%
                      (target_value,)
                    )
                target_ref = None
            return urljoin(target_base, target_ref)
        return target_value

    def get_targetvals(self):
        """
        If field description is a reference to a target type entity or field, 
        return a copy of the referenced target entity, otherwise None.
        """
        log.debug("bound_field.get_targetvals: field_description %r"%(self._field_description,))
        target_type = self._field_description.get('field_ref_type',  None)
        target_key  = self._field_description.get('field_ref_field', None)
        log.debug("bound_field.get_targetvals: target_type %s, target_key %s"%(target_type, target_key))
        if self._targetvals is None:
            if target_type:
                # Extract entity_id and type_id; default to type id from field descr
                type_id, entity_id = split_type_entity_id(self.get_field_value(), target_type)
                # Get entity type info
                #@@TODO: eliminate site param...
                coll     = self._field_description._collection
                typeinfo = EntityTypeInfo(coll.get_site(), coll, type_id)
                # Check access permission, assuming user has "VIEW" permission in current collection
                # This is primarily to prevent a loophole for accessing user account details
                #@@TODO: pass actual user permissions in to bound_field or field description or extra params
                user_permissions = ["VIEW"]
                req_permissions  = list(set( typeinfo.permissions_map[a] for a in ["view", "list"] ))
                if all([ p in user_permissions for p in req_permissions]):
                    if entity_id is None or entity_id == "":
                        raise TargetIdNotFound_Error(value=(typeinfo.type_id, self.field_name))
                    targetentity = typeinfo.get_entity(entity_id)
                    if targetentity is None:
                        raise TargetEntityNotFound_Error(value=(target_type, entity_id))
                    self._targetvals = get_entity_values(typeinfo, targetentity)
                    log.debug("bound_field.get_targetvals: %r"%(self._targetvals,))
                else:
                    log.warning(
                        "bound_field.get_targetvals: target value type %s requires %r permissions"%
                        (target_type, req_permissions)
                        )
                # # Get entity type info
                # #@@TODO: eliminate site param...
                # coll           = self._field_description._collection
                # targettypeinfo = EntityTypeInfo(coll.get_site(), coll, target_type)
                # # Check access permission, assuming user has "VIEW" permission in current collection
                # # This is primarily to prevent a loophole for accessing user account details
                # #@@TODO: pass actual user permissions in to bound_field or field description or extra params
                # user_permissions = ["VIEW"]
                # req_permissions  = list(set( targettypeinfo.permissions_map[a] for a in ["view", "list"] ))
                # if all([ p in user_permissions for p in req_permissions]):
                #     target_id    = self.get_field_value()
                #     if target_id is None or target_id == "":
                #         raise TargetIdNotFound_Error(value=(targettypeinfo.type_id, self.field_name))
                #     targetentity = targettypeinfo.get_entity(target_id)
                #     if targetentity is None:
                #         raise TargetEntityNotFound_Error(value=(targettypeinfo.type_id, target_id))
                #     self._targetvals = get_entity_values(targettypeinfo, targetentity)
                #     log.debug("bound_field.get_targetvals: %r"%(self._targetvals,))
                # else:
                #     log.warning(
                #         "bound_field.get_targetvals: target value type %s requires %r permissions"%
                #         (target_type, req_permissions)
                #         )
        log.debug("bound_field.get_targetvals: targetvals %r"%(self._targetvals,))
        return self._targetvals

    def get_link_continuation(self, link):
        """
        Return supplied base link with continuation parameter appended
        (if the link value is defined - i.e. not None or empty).
        """
        if link:
            link += self.get_continuation_param()
        return link

    def get_continuation_param(self):
        """
        Generate continuation parameter string for return back to the current request page
        """
        cparam = ""
        chere  = self.get_continuation_url()
        if chere:
            cparam = uri_params({'continuation_url': chere})
        return cparam

    def get_continuation_url(self):
        """
        Generate continuation URL for return back to the current request page
        """
        chere = ""
        if self._extras is None:
            log.warning("bound_field.get_continuation_url() - no extra context provided")
        else:
            requrl = self._extras.get("request_url", "")
            if requrl:
                chere  = uri_with_params(requrl, continuation_params(self._extras))
        log.debug('bound_field.get_continuation_url %s'%(chere,))
        return chere

    def get_field_options(self):
        #@@
        # options = self._field_description['field_choice_labels']  # OrderedDict
        # options = options.values() if options is not None else ["(no options)"]
        #@@
        options = self._field_description['field_choices']      # OrderedDict
        options = ( options.values() if options is not None else 
                    [ FieldChoice('', label="(no options)") ]
                  )
        return options

    def __getitem__(self, name):
        return self.__getattr__(name)

    def __iter__(self):
        """
        Implement iterator protocol, returning accessible value keys.
        """
        yield "entity_id"
        yield "entity_type_id"
        yield "entity_link"
        yield "entity_link_continuation"
        yield "entity_type_link"
        yield "entity_type_link_continuation"
        yield "field_edit_value"
        yield "field_view_value"
        yield "field_value"
        yield "field_value_link"
        yield "field_value_link_continuation"
        yield "target_value"
        yield "target_value_link"
        yield "target_value_link_continuation"
        yield "continuation_url"
        yield "continuation_param"
        yield "options"
        for k in self._field_description:
            yield k
        return

    def as_dict(self):
        return dict(self._field_description.items(), 
            entity=dict(self._entityvals.items()), 
            field_value=self.field_value, 
            context_extra_values=self._extras, 
            key=self._key
            )

    def __repr__(self):
        return self.fullrepr()

    def shortrepr(self):
        return (
            "bound_field(\n"+
            "  { 'key': %r\n"%(self._key,)+
            "  , 'val': %r\n"%(self.field_value,)+
            "  , 'field_description': %r\n"%(self._field_description,)+
            "  })")

    def fullrepr(self):
        return (
            ( "bound_field(\n"+
              "  { 'key': %r\n"+
              "  , 'field_edit_value': %r\n"+
              "  , 'field_view_value': %r\n"+
              "  , 'field_descr': %r\n"+
              "  , 'entity_vals': %r\n"+
              "  })\n"
            )%( self._key
              , self.field_edit_value
              , self.field_view_value 
              , self._field_description
              , dict(self._entityvals.items())
              )
            )

    def htmlrepr(self):
        return (
            "<ul>"+
            "<li>key: %s</li>"%(self._key,)+
            "<li>val: %s</li>"%(self.field_value,)+
            "<li>field_description: %r</li>"%(self._field_description,)+
            "</ul>"
            )

def get_entity_values(typeinfo=None, entity=None, entity_id=None):
    """
    Returns an entity values dictionary for a supplied entity, suitable for
    use with a bound_field object.
    """
    if not entity_id:
        entity_id = entity.get_id()
    type_id    = entity.get_type_id()
    entityvals = entity.get_values().copy()
    entityvals['entity_id']        = entity_id
    entityvals['entity_link']      = entity.get_view_url_path()
    # log.info("type_id %s"%(type_id))
    entityvals['entity_type_id']   = type_id
    #@@ typeinfo   = EntityTypeInfo(displayinfo.site, displayinfo.collection, type_id)
    if typeinfo and typeinfo.recordtype:
        entityvals['entity_type_link'] = typeinfo.recordtype.get_view_url_path()
    # This is red herring:  working version also has this
    # else:
    #     log.error("get_entity_values: No recordtype: typeinfo %r"%(typeinfo,))
    #     log.error("get_entity_values: No recordtype: entity id %s/%s"%(type_id,entity_id))
    #     log.error("get_entity_values: No recordtype: entity %r"%(entity,))
    #     traceback.print_stack()
    return entityvals

if __name__ == "__main__":
    import doctest
    doctest.testmod()

# End.



"""
Annalist tests
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import unittest
import doctest
import os.path
import shutil

from django.test import TestCase
from django.conf import settings

import annalist.util
import annalist.views.fields.render_utils
import annalist.views.fields.render_placement

from annalist.layout import Layout

test_layout     = Layout(settings.BASE_DATA_DIR)
TestBaseDir     = test_layout.SITE_PATH
TestHost        = settings.TEST_HOST            # e.g. "test.example.com"
TestBasePath    = "/" + settings.TEST_BASE_PATH # e.g. "/testsite"
TestHostUri     = settings.TEST_HOST_URI        # e.g. "http://test.example.com"
TestBaseUri     = settings.TEST_BASE_URI        # e.g. "http://test.example.com/testsite"

sitedata_target_reset = "all"

def resetSitedata(scope="all"):
    """
    Set flag to reset site and/or collection data at next test initialization
    """
    global sitedata_target_reset
    if sitedata_target_reset != "all":
        if scope != "collections":
            scope="all"
        sitedata_target_reset = scope
    return

def copySitedata(src, sitedatasrc, tgt):
    """
    Creates a set of Annalist site data by making a copy of a specified
    source tree.

    src         source directory containing site collection test data to be copied
    sitedatasrc source directory containing standard site data to be copied
    tgt         target directory in which the site data is created
    """
    # Confirm existence of target directory
    log.debug("copySitedata: src %s"%(src))
    log.debug("copySitedata: tgt %s"%(tgt))
    assert os.path.exists(src), "Check source directory (%s)"%(src)
    assert os.path.exists(sitedatasrc), "Check site data source directory (%s)"%(sitedatasrc)
    assert tgt.startswith(TestBaseDir)
    # annalist.util.replacetree(src, tgt)
    # Site data is not updated by the tests, so initialize it just once for each test suite run
    sitedatatgt = os.path.join(tgt, test_layout.SITEDATA_DIR)
    global sitedata_target_reset
    if sitedata_target_reset == "all":
        annalist.util.replacetree(src, tgt)
        for sdir in ("users", "types", "lists", "views", "groups", "fields", "enums"):
            s = os.path.join(sitedatasrc, sdir)
            d = os.path.join(sitedatatgt, sdir)
            shutil.copytree(s, d)
    elif sitedata_target_reset == "collections":
        ds = os.path.join(src, "c")
        dt = os.path.join(tgt, "c")
        annalist.util.replacetree(ds, dt)
    sitedata_target_reset = "none"
    # Confirm existence of target directories
    assert os.path.exists(tgt), "checking target directory created (%s)"%(tgt)
    assert os.path.exists(sitedatatgt), "checking target sitedata directory created (%s)"%(sitedatatgt)
    return tgt

def init_annalist_test_site():
    log.debug("init_annalist_test_site")
    copySitedata(
        settings.SITE_SRC_ROOT+"/sampledata/init/"+test_layout.SITE_DIR, 
        settings.SITE_SRC_ROOT+"/annalist/sitedata",
        TestBaseDir)
    return

def load_tests(loader, tests, ignore):
    log.debug("load_tests")
    #init_annalist_test_site()
    # See http://stackoverflow.com/questions/2380527/django-doctests-in-views-py
    if os.name == "posix":
        # The doctest stuff doesn't seem to work on Windows
        # (These add a total of 12 tests to the overall test)
        tests.addTests(doctest.DocTestSuite(annalist.util))
        tests.addTests(doctest.DocTestSuite(annalist.views.uri_builder))
        tests.addTests(doctest.DocTestSuite(annalist.views.form_utils.fieldchoice))
        tests.addTests(doctest.DocTestSuite(annalist.views.fields.render_utils))
        tests.addTests(doctest.DocTestSuite(annalist.views.fields.bound_field))
        tests.addTests(doctest.DocTestSuite(annalist.views.fields.render_placement))
        tests.addTests(doctest.DocTestSuite(annalist.models.entityfinder))
        # tests.addTests(doctest.DocTestSuite(annalist.tests.entity_testutils))
    else:
        log.warning("Skipping doctests for non-posix system")
    return tests

# Test helper functions

def unicode_to_str(u):
    return str(u) if isinstance(u, unicode) else u

def tuple_to_str(t):
    return tuple((unicode_to_str(v) for v in t))

def dict_to_str(d):
    return dict(map(tuple_to_str, d.items()))

# End.



#!/usr/bin/python

"""
Context manager for temporarily suppressing logging
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2011-2014, University of Oxford"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

class SuppressLogging:
    """
    Context handler class that suppresses logging for some controlled code.
    """
    
    def __init__(self, loglevel):
        logging.disable(loglevel)
        return
    
    def __enter__(self):
        return 

    def __exit__(self, exctype, excval, exctraceback):
        logging.disable(logging.NOTSET)
        return False

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    log.info("log this")
    with SuppressLogging(logging.WARNING):
        log.info("don't log this")
        log.warning("don't log this warning")
        log.error("log this error while up to WARNING suppressed")
    log.info("log this again")

# End.
